from typing import Any, Callable

import wrapt

import numba as nb
import numpy as np
import pandas as pd
from numba import prange
from pandas.api.types import infer_dtype

from anndata import AnnData

from cellrank import logging as logg

jit_kwargs = {"nogil": True, "cache": True, "fastmath": True}


@nb.njit(parallel=False, **jit_kwargs)
def _np_apply_along_axis(func1d, axis: int, arr: np.ndarray) -> np.ndarray:
    """Apply a reduction function over a given axis.

    Parameters
    ----------
    func1d
        Reduction function that operates only on 1 dimension.
    axis
        Axis over which to apply the reduction.
    arr
        The array to be reduced.

    Returns
    -------
    The reduced array.
    """
    assert arr.ndim == 2
    assert axis in [0, 1]

    if axis == 0:
        result = np.empty(arr.shape[1])
        for i in range(len(result)):
            result[i] = func1d(arr[:, i])
        return result

    result = np.empty(arr.shape[0])
    for i in range(len(result)):
        result[i] = func1d(arr[i, :])

    return result


@nb.njit(**jit_kwargs)
def np_mean(array: np.ndarray, axis: int) -> np.ndarray:  # noqa
    return _np_apply_along_axis(np.mean, axis, array)


@nb.njit(**jit_kwargs)
def np_std(array: np.ndarray, axis: int) -> np.ndarray:  # noqa
    return _np_apply_along_axis(np.std, axis, array)


@nb.njit(**jit_kwargs)
def norm(array: np.ndarray, axis: int) -> np.ndarray:  # noqa
    return _np_apply_along_axis(np.linalg.norm, axis, array)


# this is faster than using flat array
@nb.njit(parallel=True)
def _random_normal(
    m: np.ndarray,
    v: np.ndarray,
    n_samples: int = 1,
) -> np.ndarray:
    """Sample number from normal distribution.

    Parameters
    ----------
    m
        Mean vector.
    v
        Variance vector.
    n_samples
        Number of samples to be generated.

    Returns
    -------
    Array of shape ``(n_samples x m.shape[0])``.
    """
    assert m.ndim == 1, "Means are not 1-dimensional."
    assert m.shape == v.shape, "Means and variances have different shape."

    if n_samples == 1:
        return np.expand_dims(np.array([np.random.normal(m[i], v[i]) for i in prange(m.shape[0])]), 0)  # noqa: NPY002

    return np.array(
        [[np.random.normal(m[i], v[i]) for _ in prange(n_samples)] for i in prange(m.shape[0])]  # noqa: NPY002
    ).T


@nb.njit(**jit_kwargs)
def _calculate_starts(indptr: np.ndarray, ixs: np.ndarray) -> np.ndarray:
    """Get the position where to put the data.

    Parameters
    ----------
    indptr
        Pointer of indices from :class:`~scipy.sparse.csr_matrix`.
    ixs
        Row indices for which to calculate the starts.

    Returns
    -------
    The starting positions.
    """
    starts = np.cumsum(indptr[ixs + 1] - indptr[ixs])
    return np.hstack((np.array([0], dtype=starts.dtype), starts))


def _get_basis(adata: AnnData, basis: str) -> np.ndarray:
    try:
        return adata.obsm[f"X_{basis}"]
    except KeyError:
        try:
            return adata.obsm[basis]  # e.g. 'spatial'
        except KeyError:
            raise KeyError(f"Unable to find a basis in `adata.obsm['X_{basis}']` or `adata.obsm[{basis!r}]`.") from None


def _ensure_numeric_ordered(adata: AnnData, key: str) -> pd.Series:
    if key not in adata.obs.keys():
        raise KeyError(f"Unable to find data in `adata.obs[{key!r}]`.")

    exp_time = adata.obs[key].copy()
    if not np.issubdtype(np.asarray(exp_time).dtype, np.number):
        try:
            exp_time = np.asarray(exp_time).astype(float)
        except Exception as e:  # noqa: BLE001/Cannot interpret
            raise TypeError(
                f"Unable to convert `adata.obs[{key!r}]` of type `{infer_dtype(adata.obs[key])}` to `float`."
            ) from e

    if not isinstance(exp_time.dtype, pd.CategoricalDtype):
        logg.debug(f"Converting `adata.obs[{key!r}]` to `categorical`")
        exp_time = np.asarray(exp_time)
        categories = sorted(set(exp_time[~np.isnan(exp_time)]))
        if len(categories) > 100:  # arbitrary threshold
            raise ValueError(
                f"Unable to convert `adata.obs[{key!r}]` to `categorical` since it "
                f"would create `{len(categories)}` categories."
            )
        exp_time = pd.Series(
            pd.Categorical(
                exp_time,
                categories=categories,
                ordered=True,
            )
        )

    if not exp_time.cat.ordered:
        logg.warning("Categories are not ordered. Using ascending order")
        exp_time.cat = exp_time.cat.as_ordered()

    exp_time = pd.Series(pd.Categorical(exp_time, ordered=True), index=adata.obs_names)
    if exp_time.isnull().any():
        raise ValueError("Series contains NaN value(s).")

    n_cats = len(exp_time.cat.categories)
    if n_cats < 2:
        raise ValueError(f"Expected to find at least `2` categories, found `{n_cats}`.")

    return exp_time


@wrapt.decorator
def require_tmat(
    wrapped: Callable[..., Any],
    instance: "KernelExpression",  # noqa: F821
    args: Any,
    kwargs: Any,
) -> Any:
    """Require that the transition matrix is computed before calling the wrapped function."""
    # this can trigger combinations, but not individual kernels
    if instance.transition_matrix is None:
        raise RuntimeError("Compute transition matrix first as `.compute_transition_matrix()`.")
    return wrapped(*args, **kwargs)



from cellrank.pl._aggregate_fate_probs import aggregate_fate_probabilities
from cellrank.pl._circular_projection import circular_projection
from cellrank.pl._cluster_trends import cluster_trends
from cellrank.pl._gene_trend import gene_trends
from cellrank.pl._heatmap import heatmap
from cellrank.pl._log_odds import log_odds

__all__ = [
    "aggregate_fate_probabilities",
    "circular_projection",
    "cluster_trends",
    "gene_trends",
    "heatmap",
    "log_odds",
]



import textwrap
from typing import Any

from docrep import DocstringProcessor

__all__ = ["d", "inject_docs"]

_adata = """\
adata
    Annotated data object."""
_adata_ret = """\
:class:`anndata.AnnData`
    Annotated data object."""
_plotting = """\
figsize
    Size of the figure.
dpi
    Dots per inch.
save
    Filename where to save the plot."""
_n_jobs = """\
n_jobs
    Number of parallel jobs. If -1, use all available cores. If :obj:`None` or 1, the execution is sequential."""
_parallel = f"""\
show_progress_bar
    Whether to show a progress bar. Disabling it may slightly improve performance.
{_n_jobs}
backend
    Which backend to use for parallelization. See :class:`~joblib.Parallel` for valid options."""
_model = """\
model
    Model based on :class:`~cellrank.models.BaseModel` to fit.
    If a :class:`dict`, gene and lineage specific models can be specified. Use ``'*'`` to indicate
    all genes or lineages, for example ``{'gene_1': {'*': ...}, 'gene_2': {'lineage_1': ..., '*': ...}}``."""
_just_plots = """\
Nothing, just plots the figure. Optionally saves it based on ``save``."""
_plots_or_returns_models = """\
If ``return_models = False``, just plots the figure and optionally saves it based on ``save``.
Otherwise returns the fitted models as ``{'gene_1': {'lineage_1': <model_11>, ...}, ...}``.
Models which have failed will be instances of :class:`cellrank.models.FailedModel`."""
_backward = """\
backward
    Direction of the process."""
_eigen = """\
which
    How to sort the eigenvalues. Valid option are:

    - ``'LR'`` - the largest real part.
    - ``'LM'`` - the largest magnitude.
alpha
    Used to compute the *eigengap*. ``alpha`` is the weight given to the deviation of an eigenvalue from one."""
_n_cells = """\
n_cells
    Number of most likely cells from each macrostate to select."""
_fit = """\
n_lineages
    Number of lineages. If :obj:`None`, it will be determined automatically.
cluster_key
    Match computed states against pre-computed clusters to annotate the states.
    For this, provide a key from :attr:`anndata.AnnData.obs` where cluster labels have been computed.
keys
    Determines which states to use by passing their names. Furthermore, these states can be combined, e.g.,
    if the states are ``['Neuronal_1', 'Neuronal_1', 'OPC']``, then passing
    ``keys = ['Neuronal_1, Neuronal_2', 'OPC']`` means that the two neuronal states are treated as one."""
_density_correction = (
    "Optionally, we apply a density correction as described in :cite:`coifman:05`, "
    "where we use the implementation of :cite:`haghverdi:16`."
)
_time_range = """\
time_range
    Specify start and end times:

    - :class:`tuple` - it specifies the minimum and maximum pseudotime. Both values can be :obj:`None`,
      in which case the minimum is the earliest pseudotime and the maximum is automatically determined.
    - :class:`float` - it specifies the maximum pseudotime."""

_velocity_mode = """\
mode
    How to compute transition probabilities. Valid options are:

    - ``{m.DETERMINISTIC!r}`` - deterministic computation that doesn't propagate uncertainty.
    - ``{m.MONTE_CARLO!r}`` - Monte Carlo average of randomly sampled velocity vectors.
    - ``{m.STOCHASTIC!r}`` - second order approximation, only available when :mod:`jax` is installed."""
_velocity_backward_mode = """\
backward_mode
    Only matters if initialized as :attr:`backward = True <backward>`. Valid options are:

    - ``{b.TRANSPOSE!r}`` - compute transitions from neighboring cells :math:`j` to cell :math:`i`.
    - ``{b.NEGATE!r}`` - negate the velocity vector."""
_copy = """Return a copy of self."""
_initial = "initial"
_terminal = "terminal"
_model_callback = """\
callback
    Function which takes a :class:`~cellrank.models.BaseModel` and some keyword arguments
    for :meth:`~cellrank.models.BaseModel.prepare` and returns the prepared model.
    Can be specified in gene- and lineage-specific manner, similarly to the :attr:`model`."""
_genes = """\
genes
    Genes in :attr:`~anndata.AnnData.var_names`."""
_softmax_scale = """\
softmax_scale
    Scaling parameter for the softmax. If :obj:`None`, it will be estimated using ``1 / median(correlations)``.
    The idea behind this is to scale the softmax to counter everything tending to orthogonality in high dimensions."""
_time_mode = """\
mode
    Valid options are:

    - ``'embedding'`` - plot the embedding while coloring in continuous or categorical observations.
    - ``'time'`` - plot the pseudotime on x-axis and the probabilities/memberships on y-axis."""
_write_to_adata = """\
Updates the :attr:`adata` with the following fields:

- :attr:`obsp['{{key}}'] <anndata.AnnData.obsp>` - the transition matrix.
- :attr:`uns['{{key}}_params'] <anndata.AnnData.uns>` - parameters used for the calculation."""
_en_cutoff_p_thresh = """\
en_cutoff
    If ``cluster_key`` is given, this parameter determines when an approximate recurrent class will
    be labeled as *'Unknown'*, based on the entropy of the distribution of cells over transcriptomic clusters.
p_thresh
    If cell cycle scores were provided, a *Wilcoxon rank-sum test* is conducted to identify cell-cycle states.
    If the test returns a positive statistic and a p-value smaller than ``p_thresh``, a warning will be issued."""
_return_models = """\
return_models
    If :obj:`True`, return the fitted models for each gene in ``genes`` and lineage in ``lineages``."""
_basis = """\
basis
    Basis to use when ``mode = 'embedding'``. If :obj:`None`, use ``'umap'``."""
_velocity_scheme = """\
similarity
    Similarity measure between cells as described in :cite:`li:20`. Valid options are:

    - ``{s.CORRELATION!r}`` - :class:`~cellrank.kernels.utils.Correlation`.
    - ``{s.COSINE!r}`` - :class:`~cellrank.kernels.utils.Cosine`.
    - ``{s.DOT_PRODUCT!r}`` - :class:`~cellrank.kernels.utils.DotProduct`.

    Alternatively, any function can be passed as long as it follows the signature of
    :meth:`cellrank.kernels.utils.SimilarityABC.__call__`."""
_cond_num = """\
compute_cond_num
    Whether to compute condition number of the transition matrix. Note that this might be costly,
    since it does not use sparse implementation."""
_soft_scheme_fmt = """\
b
    The growth rate of generalized logistic function.{}
nu
    Affects near which asymptote maximum growth occurs.{}"""
_rw_ixs = """\
Can be specified as:

    - :class:`dict` - dictionary with 1 key in :attr:`~anndata.AnnData.obs` with values corresponding
      to either 1 or more clusters (if the column is categorical) or a :class:`tuple` specifying
      :math:`[min, max]` interval from which to select the indices.
    - :class:`~typing.Sequence` - sequence of cell ids in :attr:`~anndata.AnnData.obs_names`.
"""
_gene_symbols = """\
gene_symbols
    Key in :attr:`~anndata.AnnData.var` to use instead of :attr:`~anndata.AnnData.var_names`."""
_absorption_utils = """\
solver
    Solver to use for the linear problem. Options are ``'direct'``, ``'gmres'``, ``'lgmres'``, ``'bicgstab'`` or
    ``'gcrotmk'`` when ``use_petsc = False``.

    Information on the :mod:`scipy` iterative solvers can be found in :mod:`scipy.sparse.linalg` or for
    the ``petsc`` solvers `here <https://petsc.org/release/overview/linear_solve_table/>`__.
use_petsc
    Whether to use solvers from :mod:`petsc4py` or :mod:`scipy`. Recommended for large problems.
    If no installation is found, defaults to :func:`~scipy.sparse.linalg.gmres`.
n_jobs
    Number of parallel jobs to use when using an iterative solver.
backend
    Which backend to use for multiprocessing. See :class:`~joblib.Parallel` for valid options.
show_progress_bar
    Whether to show progress bar. Only used when ``solver != 'direct'``.
tol
    Convergence tolerance for the iterative solver. The default is fine for most cases, only consider
    decreasing this for severely ill-conditioned matrices.
preconditioner
    Preconditioner to use, only available when ``use_petsc = True``. For valid options, see
    `here <https://petsc.org/release/docs/manual/ksp/?highlight=pctype#preconditioners>`__.
    We recommend the ``'ilu'`` preconditioner for badly conditioned problems."""
which = """\
which
    Whether to compute initial or terminal states."""
allow_overlap = """\
allow_overlap
    Whether to allow overlapping names between initial and terminal states.
"""


def inject_docs(**kwargs: Any):
    def decorator(obj):
        obj.__doc__ = textwrap.dedent(obj.__doc__).format(**kwargs)
        return obj

    def decorator2(obj):
        obj.__doc__ = textwrap.dedent(kwargs["__doc__"])
        return obj

    if isinstance(kwargs.get("__doc__", None), str) and len(kwargs) == 1:
        return decorator2

    return decorator


d = DocstringProcessor(
    plotting=_plotting,
    n_jobs=_n_jobs,
    parallel=_parallel,
    model=_model,
    adata=_adata,
    adata_ret=_adata_ret,
    just_plots=_just_plots,
    backward=_backward,
    initial=_initial,
    terminal=_terminal,
    eigen=_eigen,
    initial_or_terminal=f"{_initial} or {_terminal}",
    n_cells=_n_cells,
    fit=_fit,
    copy=_copy,
    density_correction=_density_correction,
    time_range=_time_range,
    velocity_mode=_velocity_mode,
    velocity_backward_mode=_velocity_backward_mode,
    model_callback=_model_callback,
    genes=_genes,
    softmax_scale=_softmax_scale,
    time_mode=_time_mode,
    write_to_adata=_write_to_adata,
    en_cutoff_p_thresh=_en_cutoff_p_thresh,
    return_models=_return_models,
    plots_or_returns_models=_plots_or_returns_models,
    basis=_basis,
    velocity_scheme=_velocity_scheme,
    cond_num=_cond_num,
    soft_scheme=_soft_scheme_fmt.format("", "", ""),
    soft_scheme_kernel=_soft_scheme_fmt.format(*([" Only used when ``threshold_scheme = 'soft'``."] * 3)),
    rw_ixs=_rw_ixs,
    gene_symbols=_gene_symbols,
    absorption_utils=_absorption_utils,
    which=which,
    allow_overlap=allow_overlap,
)



import enum
from typing import Any, Optional, Sequence, Union

import numba as nb
import numpy as np
import scipy.sparse as sp
from numba import prange
from sklearn.utils.sparsefuncs import csc_median_axis_0

from anndata import AnnData

import cellrank.logging as logg
from cellrank._utils._docs import d, inject_docs
from cellrank._utils._enum import ModeEnum
from cellrank._utils._parallelize import parallelize
from cellrank._utils._utils import valuedispatch

__all__ = ["_get_offset"]

_OFFSET_KEY = "cellrank_offset"


class NormMode(ModeEnum):  # noqa: D101
    TMM = enum.auto()
    RLE = enum.auto()
    UPPER_QUANT = enum.auto()
    NONE = enum.auto()


@d.dedent
def _extract_data(data: AnnData, layer: Optional[str] = None, use_raw: bool = True) -> Union[np.ndarray, sp.spmatrix]:
    """Extract expression data from an object.

    Parameters
    ----------
    data
        Annotated data object or an array.
    layer
        Key in :attr:`~anndata.AnnData.layers` accessed when ``use_raw = False``.
        Only used when ``data`` is :class:`~anndata.AnnData`.
    use_raw
        Whether to access :attr:`anndata.AnnData.raw`. Only used when ``data`` is :class:`~anndata.AnnData`.

    Returns
    -------
    The extracted expression data.
    """
    if isinstance(data, AnnData):
        if use_raw:
            if not hasattr(data, "raw"):
                raise AttributeError("No `.raw` attribute found.")
            if data.raw is None:
                raise ValueError("Attribute `.raw` is None.")
            x = data.raw.X
        elif layer is not None:
            if layer not in data.layers:
                raise KeyError(
                    f"Layer `{layer!s}` not found in `adata.layers`. "
                    f"Valid options are: `{list(data.layers.keys())}`."
                )
            x = data.layers[layer]
        else:
            x = data.X
    elif not isinstance(data, (np.ndarray, sp.spmatrix)):
        raise TypeError(
            f"Expected parameter `data` to be either `anndata.AnnData`, `numpy.ndarray` "
            f"or `scipy.sparse.spmatrix`, found `{type(data).__name__!r}`."
        )
    else:
        x = data

    return x


@nb.njit(fastmath=True, cache=True)
def _rankdata(a: np.ndarray, method: str = "average") -> np.ndarray:
    assert method in (
        "average",
        "min",
        "max",
        "dense",
        "ordinal",
    ), "Invalid ranking method."

    arr = np.ravel(np.asarray(a))
    # algo = 'mergesort' if method == 'ordinal' else 'quicksort'
    sorter = np.argsort(arr)  # , kind=algo)

    inv = np.empty(sorter.size, dtype=np.intp)
    inv[sorter] = np.arange(sorter.size, dtype=np.intp)

    if method == "ordinal":
        return inv + 1.0

    arr = arr[sorter]
    obs = np.ones(len(arr), dtype=np.bool_)
    obs[1:] = arr[1:] != arr[:-1]
    dense = obs.cumsum()[inv]

    if method == "dense":
        return dense * 1.0

    # cumulative counts of each unique value
    count = np.append(np.nonzero(obs)[0], len(obs))

    if method == "max":
        return count[dense] * 1.0

    if method == "min":
        return count[dense - 1] + 1.0

    # average method
    return 0.5 * (count[dense] + count[dense - 1] + 1.0)


@inject_docs(m=NormMode)
def _calculate_norm_factors(
    data: Union[AnnData, np.ndarray, sp.spmatrix],
    method: str = NormMode.TMM,
    layer: Optional[str] = None,
    use_raw: bool = True,
    library_size: Optional[np.ndarray] = None,
    ref_ix: Optional[int] = None,
    logratio_trim: float = 0.3,
    sum_trim: float = 0.05,
    weight: bool = True,
    a_cutoff: float = -1e10,
    perc: float = 0.75,
) -> np.ndarray:
    """Calculate normalization factors.

    This uses the `edgeR <https://www.rdocumentation.org/packages/edgeR/versions/3.14.0/topics/calcNormFactors>`_
    implementation.

    Parameters
    ----------
    data
        Data of shape ``(n_cells, n_genes)`` containing, e.g., the counts.
    method
        Which method to use. Valid options are:

        - ``{m.TMM!r}`` - weighted trimmed mean of M-values from :cite:`robinson:10`.
        - ``{m.RLE!r}`` - relative log expression from [Anders10]_.
        - ``{m.UPPER_QUANT!r}`` - upper-quartile normalization method from [Bullard10]_.
        - ``{m.NONE!r}`` - all the factors are set to :math:`1`.
    layer
        Layer in :attr:`anndata.AnnData.layers` or `None` for :attr:`anndata.AnnData.X`.
        Only used when ``use_raw = False``.
    use_raw
        Whether to access :attr:`~anndata.AnnData.raw` or not.
    library_size
        Library size. If :obj:`None`, it will be set as the sum of values for each cell.
    ref_ix
        Index of a reference cell. If :obj:`None`, it will be determined automatically.
    logratio_trim
        Amount of trim to use on log-ratios ('M' values) when ``method = {m.TMM!r}``.
    sum_trim
        Amount of trim to use on the combined absolute levels ('A' values) when ``method = {m.TMM!r}``.
    weight
        Whether to compute asymptotic binomial precision weights when ``method = {m.TMM!r}``.
    a_cutoff
        Cutoff on 'A' values to use before trimming when ``method = {m.TMM!r}``.
    perc
        Percentile of the counts that is aligned when ``method = {m.UPPER_QUANT!r}``.

    Returns
    -------
    Array of shape ``(data.shape[0],)`` containing the factors.

    References
    ----------
    .. [Anders10] Anders, S. *et al.* (2010),
        *Differential expression analysis for sequence count data*,
        `Genome Biology <https://doi.org/10.1186/gb-2010-11-10-r106>`__.
    .. [Bullard10] Bullard, J. H. *et al.* (2010),
        *Evaluation of statistical methods for normalization and differential expression in mRNA-Seq experiments*,
        `BMC Bioinformatics <https://doi.org/10.1186/1471-2105-11-94>`__.
    """
    method = NormMode(method)

    if not (0 <= perc <= 1):
        raise ValueError(f"Expected the percentile to be within interval `[0, 1]`, found `{perc}`.")

    x = _extract_data(data, layer, use_raw)

    if library_size is None:
        library_size = np.array(x.sum(1)).squeeze()
    elif library_size.shape != (x.shape[0],):
        raise ValueError(f"Expected `library_size` to be of shape `({x.shape[0]},)`, found `{library_size.shape}`.")

    f = _dispatch_computation(
        method,
        x=x,
        library_size=library_size,
        ref_ix=ref_ix,
        logratio_trim=logratio_trim,
        sum_trim=sum_trim,
        weight=weight,
        a_cutoff=a_cutoff,
        p=perc,
    )

    return f / np.exp(np.mean(np.log(f)))


@valuedispatch
def _dispatch_computation(mode, *_args, **_kwargs):
    raise NotImplementedError(mode)


@_dispatch_computation.register(NormMode.TMM)
def _(
    x: Union[np.ndarray, sp.spmatrix],
    library_size: np.ndarray,
    ref_ix: Optional[int] = None,
    **kwargs,
) -> np.ndarray:
    if ref_ix is None:
        f75 = _calc_factor_quant(x, library_size=library_size, p=0.75)
        ref_ix = np.argmin(np.abs(f75 - np.mean(f75)))

    ref = x[ref_ix]
    if sp.issparse(ref):
        ref = ref.A.squeeze(0)  # (genes,)

    return parallelize(
        _calc_factor_weighted,
        collection=np.arange(x.shape[0]),
        show_progress_bar=False,
        as_array=False,
        extractor=np.concatenate,
        backend="threading",
        n_jobs=4,
    )(
        obs_=x,
        obs_lib_size_=library_size,
        ref=ref,
        ref_lib_size=library_size[ref_ix],
        **kwargs,
    )


@_dispatch_computation.register(NormMode.UPPER_QUANT)
def _calc_factor_quant(x: Union[np.ndarray, sp.spmatrix], library_size: np.ndarray, p: float, **_) -> np.ndarray:
    # unused, because we fix the reference cell to 0

    library_size = np.array(library_size).reshape((-1, 1))
    if not sp.issparse(x):
        # this is same as below, which is R's default
        # return mquantiles(x / library_size, prob=q, alphap=1, betap=1, axis=1).data.squeeze()
        return np.quantile(x / library_size, p, axis=1)

    y = x.multiply(1.0 / library_size).tocsr()
    return _calc_factor_quant_sparse_helper(y.data, y.indptr, p=p, n_cols=y.shape[1])


@nb.njit(parallel=True, fastmath=True, cache=True)
def _calc_factor_quant_sparse_helper(
    data: np.ndarray,
    indptr: np.ndarray,
    p: float,
    n_cols: int,
) -> np.ndarray:
    n = len(indptr) - 1
    out = np.empty(n)

    # maybe disable parallel
    for ix in prange(n):
        d = data[indptr[ix] : indptr[ix + 1]]
        tmp = np.zeros(shape=(n_cols,))
        tmp[: len(d)] = d
        out[ix] = np.quantile(tmp, p)

    return out


@_dispatch_computation.register(NormMode.RLE)
def _(x: Union[np.ndarray, sp.spmatrix], **_) -> np.ndarray:
    # unused

    # log 0 -> inf -> masked out anyway, so we select only genes in every cell
    # this is the exact replica an in edgeR
    mask = np.array((x > 0).sum(0)).squeeze() == x.shape[0]
    tmp = x[:, mask]
    if sp.issparse(tmp):
        tmp = tmp.A

    gm = np.array(np.exp(np.mean(np.log(tmp), axis=0))).squeeze()
    gm_mask = (gm > 0) & np.isfinite(gm)

    if not sp.issparse(x):
        return np.median(x[:, mask][:, gm_mask] / gm[gm_mask], axis=1)

    return csc_median_axis_0(x.tocsr()[:, mask][:, gm_mask].multiply(1.0 / gm[gm_mask]).tocsr().T)


@_dispatch_computation.register(NormMode.NONE)
def _(x: Union[np.ndarray, sp.spmatrix], **_) -> np.ndarray:
    # unused
    return np.ones((x.shape[0],), dtype=x.dtype)


def _calc_factor_weighted(
    ixs: np.ndarray,
    obs_: Union[np.ndarray, sp.spmatrix],
    obs_lib_size_: np.ndarray,
    ref: np.ndarray,
    ref_lib_size: float,
    logratio_trim: float = 0.3,
    sum_trim: float = 0.05,
    weight: bool = True,
    a_cutoff: float = -1e10,
    queue=None,
    **_: Any,
) -> np.ndarray:
    with np.errstate(divide="ignore", invalid="ignore"):
        log_ref = np.log2(ref / ref_lib_size)
    res = np.empty(shape=(len(ixs),))

    for i, ix in enumerate(ixs):
        if queue is not None:
            queue.put(1)

        obs = obs_[ix]
        if sp.issparse(obs):
            obs = obs.A.squeeze(0)  # (genes,)
        obs_lib_size = obs_lib_size_[ix]

        res[i] = _calc_factor_weighted_helper(
            obs=obs,
            obs_lib_size=obs_lib_size,
            ref=ref,
            ref_lib_size=ref_lib_size,
            log_ref=log_ref,
            a_cutoff=a_cutoff,
            logratio_trim=logratio_trim,
            sum_trim=sum_trim,
            weight=weight,
        )

    if queue is not None:
        queue.put(None)

    return res


@nb.njit(fastmath=True, cache=True)
def _calc_factor_weighted_helper(
    obs: np.ndarray,
    obs_lib_size: float,
    ref: np.ndarray,
    ref_lib_size: float,
    log_ref: np.ndarray,
    a_cutoff: float,
    logratio_trim: float,
    sum_trim: float,
    weight: bool,
) -> float:
    o_scaled = obs / obs_lib_size
    log_obs = np.log2(o_scaled)

    log_r = log_obs - log_ref
    abs_e = (log_obs + log_ref) / 2.0
    v = ((obs_lib_size - obs) / obs_lib_size / obs) + (ref_lib_size - ref) / ref_lib_size / ref

    mask = np.isfinite(log_r) & np.isfinite(abs_e) & (abs_e > a_cutoff)

    log_r = log_r[mask]
    abs_e = abs_e[mask]
    v = v[mask]

    if not len(log_r) or np.max(np.abs(log_r)) < 1e-6:
        return 1.0

    n = len(log_r)
    lol = np.floor(n * logratio_trim) + 1.0
    hil = n + 1 - lol

    los = np.floor(n * sum_trim) + 1.0
    his = n + 1 - los

    rank_log_r, rank_abs_e = _rankdata(log_r), _rankdata(abs_e)
    mask = (rank_log_r >= lol) & (rank_log_r <= hil) & (rank_abs_e >= los) & (rank_abs_e <= his)

    f = (np.nansum(log_r[mask] / v[mask]) / np.nansum(1.0 / v[mask])) if weight else np.nanmean(log_r[mask])

    return 2 ** f if np.isfinite(f) else 1.0


def _get_knotlocs(
    pseudotime: Union[np.ndarray, Sequence],
    n_knots: int,
    uniform: bool = False,
) -> np.ndarray:
    """Find knot locations.

    The first and last knots are always placed at the beginning and the of the ``pseudotime``.

    Parameters
    ----------
    pseudotime
        Pseudotemporal ordering of cells.
    n_knots
        Number of knots.
    uniform
        Whether to place the knots uniformly across the ``pseudotime`` or place them based on ``pseudotime``'s density.

    Returns
    -------
    Array of shape ``(n_knots,)`` containing the locations of knots along the pseudotime.
    """
    if n_knots <= 0:
        raise ValueError(f"Expected number of knots to be positive, found `{n_knots}`.")

    if np.any(~np.isfinite(pseudotime)):
        raise ValueError("Not all pseudotime values are finite.")

    pseudotime = np.asarray(pseudotime)

    unique = np.unique(pseudotime)
    if len(unique) in (0, 1):
        raise ValueError(f"All pseudotime values are the same: `{unique}`.")

    if pseudotime.ndim == 2 and pseudotime.shape[1] == 1:
        pseudotime = pseudotime.squeeze(1)
    if pseudotime.ndim != 1:
        raise ValueError(f"Expected `pseudotime` to have `1` dimension, found `{pseudotime.ndim}`.")

    if uniform:
        # replicate the result from not uniform
        return (
            np.linspace(np.min(pseudotime), np.max(pseudotime), n_knots, endpoint=True)
            if n_knots > 1
            else np.array([np.max(pseudotime)])
        )

    x = np.quantile(pseudotime, q=np.arange(n_knots, dtype=np.float64) / max(n_knots - 1, 1))
    x[0] = np.min(pseudotime)
    x[-1] = np.max(pseudotime)

    u, ix, c = np.unique(x, return_index=True, return_counts=True)

    if len(u) != len(x):
        knotlocs = []
        for start, end, size in zip(x[ix], x[ix[1:]], c):
            knotlocs.extend(np.linspace(start, end, size, endpoint=False))
        knotlocs.extend(np.linspace(knotlocs[-1], x[ix[-1]], c[-1] + 1, endpoint=True)[1:])
        knotlocs = np.array(knotlocs)
    else:
        knotlocs = x

    logg.debug(f"Setting knot locations to `{list(knotlocs)}`")

    return knotlocs


@inject_docs(key=_OFFSET_KEY)
@d.dedent
def _get_offset(
    adata: Union[AnnData, np.ndarray, sp.spmatrix],
    layer: Optional[str] = None,
    use_raw: bool = True,
    ref_ix: Optional[int] = None,
    recompute: bool = False,
    **kwargs,
) -> np.ndarray:
    """Return an offset for GAM.

    Parameters
    ----------
    %(adata)s
    layer
        Layer in :attr:`~anndata.AnnData.layers`` or :obj:`None` for :attr:`~anndata.AnnData.X`.
        Only used when ``use_raw=False``.
    use_raw
        Whether to access :attr:`~anndata.AnnData.raw`.
    ref_ix
        Index of a reference cell. If :obj:`None`, it will be determined automatically.
    recompute
        Whether to recompute the offset if already found in :attr:`adata.obs['{key}'] <anndata.AnnData.obs>`.
    kwargs
        Keyword arguments for :func:`_utils._calc_norm_factors`.

    Returns
    -------
    Array of shape ``(adata.n_obs,)`` containing the offset.
    """
    with np.errstate(divide="ignore", invalid="ignore"):
        if not recompute and isinstance(adata, AnnData) and _OFFSET_KEY in adata.obs:
            logg.debug(f"Fetching offset from `adata.obs[{_OFFSET_KEY!r}]`")
            return adata.obs[_OFFSET_KEY].values.copy()

        logg.debug(f"Calculating offset for `{adata.shape[0]}` cells")

        data = _extract_data(adata, layer=layer, use_raw=use_raw)
        try:
            nf = _calculate_norm_factors(adata, layer=layer, use_raw=use_raw, ref_ix=ref_ix, **kwargs)
        except Exception as e:  # noqa: BLE001
            logg.debug(f"Unable to calculate the normalization factors, setting them to `1`. Reason: `{e}`")
            nf = np.ones(len(adata), dtype=np.float64)

        offset = np.log(nf * np.array(data.sum(1)).squeeze())
        offset[offset == 0] = 1.0

        mask = ~np.isfinite(offset) | np.isnan(offset)
        if np.any(mask):
            logg.warning(f"`{np.sum(mask)}` elements are not finite. Setting them to `1`")
            offset[mask] = 1.0

        if isinstance(adata, AnnData):
            adata.obs[_OFFSET_KEY] = offset

        return offset



import collections
import enum
import math
import os
import pathlib
from typing import Any, Dict, List, Literal, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from scipy.ndimage.filters import convolve

import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib import cm, colors
from matplotlib.ticker import FormatStrFormatter
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils import Lineage
from cellrank._utils._docs import d, inject_docs
from cellrank._utils._enum import DEFAULT_BACKEND, Backend_t, ModeEnum
from cellrank._utils._parallelize import _get_n_cores
from cellrank._utils._utils import (
    _check_collection,
    _genesymbols,
    _min_max_scale,
    _unique_order_preserving,
    save_fig,
    valuedispatch,
)
from cellrank.pl._utils import (
    _callback_type,
    _create_callbacks,
    _create_models,
    _fit_bulk,
    _get_backend,
    _get_categorical_colors,
    _input_model_type,
    _return_model_type,
    _time_range_type,
)

__all__ = ["heatmap"]

_N_XTICKS = 10


class HeatmapMode(ModeEnum):
    GENES = enum.auto()
    LINEAGES = enum.auto()


@d.dedent
@inject_docs(m=HeatmapMode)
@_genesymbols
def heatmap(
    adata: AnnData,
    model: _input_model_type,
    genes: Sequence[str],
    time_key: str,
    lineages: Optional[Union[str, Sequence[str]]] = None,
    backward: bool = False,
    mode: Literal["genes", "lineages"] = HeatmapMode.LINEAGES,
    time_range: Optional[Union[_time_range_type, List[_time_range_type]]] = None,
    callback: _callback_type = None,
    cluster_key: Optional[Union[str, Sequence[str]]] = None,
    show_fate_probabilities: bool = False,
    cluster_genes: bool = False,
    keep_gene_order: bool = False,
    scale: bool = True,
    n_convolve: Optional[int] = 5,
    show_all_genes: bool = False,
    cbar: bool = True,
    lineage_height: float = 0.33,
    fontsize: Optional[float] = None,
    xlabel: Optional[str] = None,
    title: Optional[str] = None,
    cmap: colors.ListedColormap = cm.viridis,
    dendrogram: bool = True,
    return_genes: bool = False,
    return_models: bool = False,
    return_figure: bool = False,
    n_jobs: Optional[int] = 1,
    backend: Backend_t = DEFAULT_BACKEND,
    show_progress_bar: bool = True,
    figsize: Optional[Tuple[float, float]] = None,
    dpi: Optional[int] = None,
    save: Optional[Union[str, pathlib.Path]] = None,
    gene_order: Optional[Sequence[str]] = None,
    **kwargs: Any,
) -> Optional[Union[Dict[str, pd.DataFrame], Tuple[_return_model_type, Dict[str, pd.DataFrame]]]]:
    """Plot a heatmap of smoothed gene expression along specified lineages.

    .. seealso::
        - See :doc:`../../../notebooks/tutorials/estimators/800_gene_trends` on how to
          visualize the gene trends.

    This requires a pseudotemporal ordering of cellular dynamics, computed using a method like
    DPT :cite:`haghverdi:16` or Palantir :cite:`setty:19`. The function combines the pseudotemporal
    ordering with CellRank's fate probabilities to visualize each gene's expression along specific
    trajectories. In the heatmap, genes are ordered according to their peak in pseudotime, which
    emphasizes expression cascades of sequential activation.

    Parameters
    ----------
    %(adata)s
    %(model)s
    %(genes)s
    time_key
        Key in :attr:`~anndata.AnnData.obs` where the pseudotime is stored.
    lineages
        Names of the lineages for which to plot. If :obj:`None`, plot all lineages.
    %(backward)s
    mode
        Valid options are:

        - ``{m.LINEAGES!r}`` - group by ``genes`` for each lineage in ``lineages``.
        - ``{m.GENES!r}`` - group by ``lineages`` for each gene in ``genes``.
    %(time_range)s
        This can also be specified on per-lineage basis.
    %(gene_symbols)s
    %(model_callback)s
    cluster_key
        Key in :attr:`~anndata.AnnData.obs` containing categorical observations to be plotted on the top of heatmap.
        Only available when ``mode = {m.LINEAGES!r}``.
    show_fate_probabilities
        Whether to also plot fate probabilities alongside the smoothed expression.
        Only available when ``mode = {m.LINEAGES!r}``.
    cluster_genes
        Whether to cluster genes using :func:`~seaborn.clustermap` when ``mode = 'lineages'``.
    keep_gene_order
        Whether to keep the gene order for later lineages after the first was sorted.
        Only available when ``cluster_genes = False`` and ``mode = {m.LINEAGES!r}``.
    scale
        Whether to normalize the gene expression to :math:`[0, 1]`.
    n_convolve
        Size of the convolution window when smoothing fate probabilities.
    show_all_genes
        Whether to show all genes on y-axis.
    cbar
        Whether to show the colorbar.
    lineage_height
        Height of a bar when ``mode = {m.GENES!r}``.
    fontsize
        Size of the title's font.
    xlabel
        Label on the x-axis. If :obj:`None`, it is determined based on ``time_key``.
    title
        Title of the figure.
    cmap
        Colormap to use when visualizing the smoothed expression.
    dendrogram
        Whether to show dendrogram when ``cluster_genes = True``.
    return_genes
        Whether to return the sorted or clustered genes. Only available when ``mode = {m.LINEAGES!r}``.
    %(return_models)s
    return_figure
        Whether to return the figure object. Sets ``return_genes = True``
    %(parallel)s
    %(plotting)s
    kwargs
        Keyword arguments for :meth:`~cellrank.models.BaseModel.prepare`.

    Returns
    -------
    %(plots_or_returns_models)s

    - If ``return_genes = True`` and ``mode = {m.LINEAGES!r}``, returns a :class:`~pandas.DataFrame`
    containing the clustered or sorted genes.
    - If ``return_figure = True``, returns a tuple containing the figure and genes.
    """

    def find_indices(series: pd.Series, values) -> List[int]:
        def find_nearest(array: np.ndarray, value: float) -> int:
            ix = np.searchsorted(array, value, side="left")
            if ix > 0 and (ix == len(array) or math.fabs(value - array[ix - 1]) < math.fabs(value - array[ix])):
                return int(ix - 1)
            return int(ix)

        series = series.sort_values(ascending=True)
        return list(series.iloc[[find_nearest(series.values, v) for v in values]].index)

    def subset_lineage(lname: str, rng: np.ndarray) -> np.ndarray:
        time_series = adata.obs[time_key]
        ixs = find_indices(time_series, rng)
        lin = Lineage.from_adata(adata[ixs], backward=backward)
        lin = lin[lname].X.squeeze(1).copy()
        if n_convolve is None:
            return lin.copy()
        return convolve(lin, np.ones(n_convolve) / n_convolve, mode="nearest")

    def create_col_colors(lname: str, rng: np.ndarray) -> Tuple[np.ndarray, colors.Colormap, colors.Normalize]:
        color = probs[lname].colors[0]
        lin = subset_lineage(lname, rng)

        h, _, v = colors.rgb_to_hsv(colors.to_rgb(color))
        end_color = colors.hsv_to_rgb([h, 1, v])

        lineage_cmap = colors.LinearSegmentedColormap.from_list("lineage_cmap", ["#ffffff", end_color], N=len(rng))
        norm = colors.Normalize(vmin=np.min(lin), vmax=np.max(lin))
        scalar_map = cm.ScalarMappable(cmap=lineage_cmap, norm=norm)

        return (
            np.array([colors.to_hex(c) for c in scalar_map.to_rgba(lin)]),
            lineage_cmap,
            norm,
        )

    def create_col_categorical_color(cluster_key: str, rng: np.ndarray) -> np.ndarray:
        cols, mapper = _get_categorical_colors(adata, cluster_key)
        ixs = find_indices(adata.obs[time_key], rng)

        return np.array([colors.to_hex(mapper[v]) for v in adata[ixs, :].obs[cluster_key].values])

    def create_cbar(
        ax,
        x_delta: float,
        cmap: colors.Colormap,
        norm: colors.Normalize,
        label: Optional[str] = None,
    ) -> plt.Axes:
        cax = inset_axes(
            ax,
            width="1%",
            height="100%",
            loc="lower right",
            bbox_to_anchor=(x_delta, 0, 1, 1),
            bbox_transform=ax.transAxes,
        )

        _ = mpl.colorbar.ColorbarBase(
            cax,
            cmap=cmap,
            norm=norm,
            label=label,
            ticks=np.linspace(norm.vmin, norm.vmax, 5),
        )

        return cax

    @valuedispatch
    def _plot_heatmap(mode: HeatmapMode) -> plt.Figure:
        raise NotImplementedError(mode.value)

    @_plot_heatmap.register(HeatmapMode.GENES)
    def _() -> Tuple[plt.Figure, None]:
        def color_fill_rec(ax, xs, y1, y2, colors=None, cmap=cmap, **kwargs) -> None:
            colors = colors if cmap is None else cmap(colors)

            x = 0
            for i, (color, x, y1, y2) in enumerate(zip(colors, xs, y1, y2)):  # noqa: B020
                dx = (xs[i + 1] - xs[i]) if i < len(x) else (xs[-1] - xs[-2])
                ax.add_patch(plt.Rectangle((x, y1), dx, y2 - y1, color=color, ec=color, **kwargs))

            ax.plot(x, y2, lw=0)

        fig, axes = plt.subplots(
            nrows=len(genes) + show_fate_probabilities,
            figsize=(12, len(genes) + len(lineages) * lineage_height) if figsize is None else figsize,
            dpi=dpi,
            constrained_layout=True,
        )
        axes = np.ravel([axes])

        if show_fate_probabilities:
            data["fate probability"] = data[next(iter(data.keys()))]

        for ax, (gene, models) in zip(axes, data.items()):
            if scale:
                vmin, vmax = 0, 1
            else:
                c = np.array([m.y_test for m in models.values()])
                vmin, vmax = np.nanmin(c), np.nanmax(c)

            norm = colors.Normalize(vmin=vmin, vmax=vmax)

            ix = 0
            ys = [ix]

            if gene == "fate probability":
                norm = colors.Normalize(vmin=0, vmax=1)
                for ln, x in ((ln, m.x_test) for ln, m in models.items()):
                    y = np.ones_like(x)
                    c = subset_lineage(ln, x.squeeze())

                    color_fill_rec(ax, x, y * ix, y * (ix + lineage_height), colors=norm(c))

                    ix += lineage_height
                    ys.append(ix)
            else:
                for x, c in ((m.x_test, m.y_test) for m in models.values()):
                    y = np.ones_like(x)
                    c = _min_max_scale(c) if scale else c

                    color_fill_rec(ax, x, y * ix, y * (ix + lineage_height), colors=norm(c))

                    ix += lineage_height
                    ys.append(ix)

            xs = np.array([m.x_test for m in models.values()])
            x_min, x_max = np.min(xs), np.max(xs)
            ax.set_xticks(np.linspace(x_min, x_max, _N_XTICKS))

            ax.set_yticks(np.array(ys[:-1]) + lineage_height / 2)
            ax.spines["left"].set_position(("data", 0))  # move the left spine to the rectangles to get nicer yticks
            ax.set_yticklabels(models.keys(), ha="right")

            ax.set_title(gene, fontdict={"fontsize": fontsize})
            ax.set_ylabel("lineage")

            for pos in ["top", "bottom", "left", "right"]:
                ax.spines[pos].set_visible(False)

            if cbar:
                cax, _ = mpl.colorbar.make_axes(ax)
                _ = mpl.colorbar.ColorbarBase(
                    cax,
                    ticks=np.linspace(vmin, vmax, 5),
                    norm=norm,
                    cmap=cmap,
                    label="value" if gene == "fate probability" else "scaled expression" if scale else "expression",
                )

            ax.tick_params(
                top=False,
                bottom=False,
                left=True,
                right=False,
                labelleft=True,
                labelbottom=False,
            )

        ax.xaxis.set_major_formatter(FormatStrFormatter("%.3f"))
        ax.tick_params(
            top=False,
            bottom=True,
            left=True,
            right=False,
            labelleft=True,
            labelbottom=True,
        )
        ax.set_xlabel(xlabel)
        if title is not None:
            ax.set_title(title)

        return fig, None

    @_plot_heatmap.register(HeatmapMode.LINEAGES)
    def _(gene_order: Optional[Sequence[str]] = None) -> Tuple[List[plt.Figure], pd.DataFrame]:
        data_t = collections.defaultdict(dict)  # transpose
        for gene, lns in data.items():
            for ln, y in lns.items():
                data_t[ln][gene] = y

        figs = []
        sorted_genes = pd.DataFrame() if return_genes else None

        for lname, models in data_t.items():
            xs = np.array([m.x_test for m in models.values()])
            x_min, x_max = np.nanmin(xs), np.nanmax(xs)

            df = pd.DataFrame([m.y_test for m in models.values()], index=models.keys())
            df.index.name = "genes"

            if not cluster_genes:
                if gene_order is not None:
                    gene_order = [gene for gene in gene_order if gene in df.index]
                    df = df.loc[gene_order]
                else:
                    max_sort = np.argsort(np.argmax(df.apply(_min_max_scale, axis=1).values, axis=1))
                    df = df.iloc[max_sort, :]
                    if keep_gene_order:
                        gene_order = df.index

            cat_colors = None
            if cluster_key is not None:
                cat_colors = np.stack(
                    [create_col_categorical_color(c, np.linspace(x_min, x_max, df.shape[1])) for c in cluster_key],
                    axis=0,
                )

            if show_fate_probabilities:
                col_colors, col_cmap, col_norm = create_col_colors(lname, np.linspace(x_min, x_max, df.shape[1]))
                if cat_colors is not None:
                    col_colors = np.vstack([cat_colors, col_colors[None, :]])
            else:
                col_colors, col_cmap, col_norm = cat_colors, None, None

            row_cluster = cluster_genes and df.shape[0] > 1
            show_clust = row_cluster and dendrogram

            g = sns.clustermap(
                data=df,
                cmap=cmap,
                figsize=(10, min(len(genes) / 8 + 1, 10)) if figsize is None else figsize,
                xticklabels=False,
                row_cluster=row_cluster,
                col_colors=col_colors,
                colors_ratio=0,
                col_cluster=False,
                cbar_pos=None,
                yticklabels=show_all_genes or "auto",
                standard_scale=0 if scale else None,
            )

            if cbar:
                cax = create_cbar(
                    g.ax_heatmap,
                    0.1,
                    cmap=cmap,
                    norm=colors.Normalize(
                        vmin=0 if scale else np.min(df.values),
                        vmax=1 if scale else np.max(df.values),
                    ),
                    label="scaled expression" if scale else "expression",
                )
                g.fig.add_axes(cax)

                if col_cmap is not None and col_norm is not None:
                    cax = create_cbar(
                        g.ax_heatmap,
                        0.25,
                        cmap=col_cmap,
                        norm=col_norm,
                        label="fate probability",
                    )
                    g.fig.add_axes(cax)

            if g.ax_col_colors:
                main_bbox = _get_ax_bbox(g.fig, g.ax_heatmap)
                n_bars = show_fate_probabilities + (len(cluster_key) if cluster_key is not None else 0)
                _set_ax_height_to_cm(
                    g.fig,
                    g.ax_col_colors,
                    height=min(5, max(n_bars * main_bbox.height / len(df), 0.25 * n_bars)),
                )
                g.ax_col_colors.set_title(lname, fontdict={"fontsize": fontsize})
            else:
                g.ax_heatmap.set_title(lname, fontdict={"fontsize": fontsize})

            g.ax_col_dendrogram.set_visible(False)  # gets rid of top free space
            g.ax_heatmap.yaxis.tick_left()
            g.ax_heatmap.yaxis.set_label_position("right")
            # fmt: off
            g.ax_heatmap.set_xlabel(xlabel)
            g.ax_heatmap.set_xticks(np.linspace(0, len(df.columns), _N_XTICKS))
            g.ax_heatmap.set_xticklabels([round(n, 3) for n in np.linspace(x_min, x_max, _N_XTICKS)])
            if title is not None:
                g.ax_heatmap.set_title(title)
            # fmt: on
            if show_clust:
                # robustly show dendrogram, because gene names can be long
                g.ax_row_dendrogram.set_visible(True)
                dendro_box = g.ax_row_dendrogram.get_position()
                pad = 0.005
                bb = g.ax_heatmap.yaxis.get_tightbbox(g.fig.canvas.get_renderer()).transformed(
                    g.fig.transFigure.inverted()
                )
                dendro_box.x0 = bb.x0 - dendro_box.width - pad
                dendro_box.x1 = bb.x0 - pad
                g.ax_row_dendrogram.set_position(dendro_box)
            else:
                g.ax_row_dendrogram.set_visible(False)

            if return_genes:
                sorted_genes[lname] = (
                    df.index[g.dendrogram_row.reordered_ind]
                    if hasattr(g, "dendrogram_row") and g.dendrogram_row is not None
                    else df.index
                )

            figs.append(g)

        return figs, sorted_genes

    mode = HeatmapMode(mode)

    probs = Lineage.from_adata(adata, backward=backward)
    if lineages is None:
        lineages = probs.names
    elif isinstance(lineages, str):
        lineages = [lineages]
    lineages = _unique_order_preserving(lineages)
    probs = probs[lineages]

    if cluster_key is not None:
        if isinstance(cluster_key, str):
            cluster_key = [cluster_key]
        cluster_key = _unique_order_preserving(cluster_key)

    if isinstance(genes, str):
        genes = [genes]
    genes = _unique_order_preserving(genes)
    _check_collection(adata, genes, "var_names", use_raw=kwargs.get("use_raw", False))

    kwargs["backward"] = backward
    kwargs["time_key"] = time_key
    models = _create_models(model, genes, lineages)
    all_models, data, genes, lineages = _fit_bulk(
        models,
        _create_callbacks(adata, callback, genes, lineages, **kwargs),
        genes,
        lineages,
        time_range,
        return_models=True,  # always return (better error messages)
        filter_all_failed=True,
        parallel_kwargs={
            "show_progress_bar": show_progress_bar,
            "n_jobs": _get_n_cores(n_jobs, len(genes)),
            "backend": _get_backend(models, backend),
        },
        **kwargs,
    )

    xlabel = time_key if xlabel is None else xlabel

    logg.debug(f"Plotting `{mode!r}` heatmap")
    if gene_order is not None and mode != HeatmapMode.LINEAGES:
        gene_order = [gene for gene in gene_order if gene in genes]

    if mode == HeatmapMode.LINEAGES:
        fig, genes = _plot_heatmap(mode, gene_order)
    else:
        fig, genes = _plot_heatmap(mode)

    if return_figure:
        return fig, genes

    if save is not None and fig is not None:
        if isinstance(fig, plt.Figure):
            save_fig(fig, save)
        elif len(fig) == 1:
            save_fig(fig[0], save)
        else:
            for ln, f in zip(lineages, fig):
                save_fig(f, os.path.join(save, f"lineage_{ln}"))

    if return_genes and mode == HeatmapMode.LINEAGES:
        return (all_models, genes) if return_models else genes
    if return_models:
        return all_models


def _get_ax_bbox(fig: plt.Figure, ax: plt.Axes):
    return ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())


def _set_ax_height_to_cm(fig: plt.Figure, ax: plt.Axes, height: float) -> None:
    from mpl_toolkits.axes_grid1 import Divider, Size

    height /= 2.54  # cm to inches

    bbox = _get_ax_bbox(fig, ax)

    hori = [Size.Fixed(bbox.x0), Size.Fixed(bbox.width), Size.Fixed(bbox.x1)]
    vert = [Size.Fixed(bbox.y0), Size.Fixed(height), Size.Fixed(bbox.y1)]

    divider = Divider(fig, (0.0, 0.0, 1.0, 1.0), hori, vert, aspect=False)

    ax.set_axes_locator(divider.new_locator(nx=1, ny=1))



from typing import Any, Dict, List, Literal, Optional, Sequence, Union

import numpy as np
import pandas as pd
import scipy.stats as st

from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils._docs import d
from cellrank._utils._utils import (
    _cluster_X,
    _complex_warning,
    _filter_cells,
    _get_connectivities,
)
from cellrank.estimators.mixins import EigenMixin, LinDriversMixin
from cellrank.estimators.terminal_states._term_states_estimator import (
    TermStatesEstimator,
)
from cellrank.kernels._utils import _get_basis

__all__ = ["CFLARE"]


@d.dedent
class CFLARE(TermStatesEstimator, LinDriversMixin, EigenMixin):
    """Clustering and filtering of left and right eigenvectors (CFLARE).

    This estimator computes initial and terminal states, as well as fate probabilities.
    It uses the left eigenvectors of the transition matrix to filter to a set of (approximately)
    recurrent cells and the right eigenvectors to cluster this set of cells into discrete groups.
    CFLARE exists mostly for legacy reasons, we recommend using the :class:`cellrank.estimators.GPCCA`
    estimator, which is mathematically more principled.

    Parameters
    ----------
    %(base_estimator.parameters)s
    """

    @d.dedent
    def fit(self, k: int = 20, **kwargs: Any) -> "TermStatesEstimator":
        """Prepare self for terminal states prediction.

        Parameters
        ----------
        k
            Number of eigenvectors to compute.
        kwargs
            Keyword arguments for :meth:`compute_eigendecomposition`.

        Returns
        -------
        Returns self and updates the following fields:

        - :attr:`eigendecomposition` - %(eigen.summary)s
        """
        return self.compute_eigendecomposition(k=k, only_evals=False, **kwargs)

    @d.dedent
    def predict(
        self,
        use: Optional[Union[int, Sequence[int]]] = None,
        percentile: Optional[int] = 98,
        method: Literal["leiden", "kmeans"] = "leiden",
        cluster_key: Optional[str] = None,
        n_clusters_kmeans: Optional[int] = None,
        n_neighbors: int = 20,
        resolution: float = 0.1,
        n_matches_min: int = 0,
        n_neighbors_filtering: int = 15,
        basis: Optional[str] = None,
        n_comps: int = 5,
        scale: Optional[bool] = None,
    ) -> "CFLARE":
        """Find approximate recurrent classes of the Markov chain.

        Filter to obtain recurrent states from left eigenvectors.
        Cluster to obtain approximate recurrent classes from right eigenvectors.

        Parameters
        ----------
        use
            Which or how many first eigenvectors to use as features for filtering and clustering.
            If :obj:`None`, use the *eigengap* statistic.
        percentile
            Threshold used for filtering out cells which are most likely transient states. Cells which are in the
            lower ``percentile`` percent of each eigenvector will be removed from the data matrix.
        method
            Method to be used for clustering. Valid option are:

            - ``'kmeans'`` - :class:`~sklearn.cluster.KMeans`.
            - ``'leiden'`` - :func:`~scanpy.tl.leiden`.
        cluster_key
            Key in :attr:`~anndata.AnnData.obs` in order to associate names and colors with :attr:`terminal_states`.
        n_clusters_kmeans
            Number of clusters when ``method = 'kmeans'``. If :obj:`None`, this is set to ``use + 1``.
        n_neighbors
            Number of neighbors in a kNN graph. This is the :math:`K` parameter for that,
            the number of neighbors for each cell. Only used when ``method = 'leiden'``.
        resolution
            Resolution parameter for :func:`~scanpy.tl.leiden`. Should be chosen relatively small.
        n_matches_min
            Filters out cells which don't have at least ``n_matches_min`` neighbors from the same category.
            This filters out some cells which are transient but have been misassigned.
        n_neighbors_filtering
            Parameter for filtering cells. Cells are filtered out if they don't have at least ``n_matches_min``
            neighbors among the ``n_neighbors_filtering`` nearest cells.
        basis
            Key from :attr:`~anndata.AnnData.obsm` as additional features for clustering.
            If `None`, use only the right eigenvectors.
        n_comps
            Number of embedding components to be used when ``basis != None``.
        scale
            Scale the values to z-scores. If :obj:`None`, scale the values if ``basis != None``.

        Returns
        -------
        Returns self and just updates the following fields:

        - :attr:`terminal_states` - %(tse_term_states.summary)s
        - :attr:`terminal_states_probabilities` - %(tse_term_states_probs.summary)s
        """

        def convert_use(use: Optional[Union[int, Sequence[int], np.ndarray]]) -> List[int]:
            if method not in ["kmeans", "leiden"]:
                raise ValueError(f"Invalid method `{method!r}`. Valid options are `leiden` or `kmeans`.")

            if use is None:
                use = eig["eigengap"] + 1  # add one b/c indexing starts at 0
            if isinstance(use, int):
                use = list(range(use))
            elif not isinstance(use, (np.ndarray, Sequence)):
                raise TypeError(f"Expected `use` to be `int` or a `Sequence`, found `{type(use).__name__}`.")
            use = list(use)

            if not use:
                raise ValueError("No eigenvectors have been selected.")

            muse, mevecs = max(use), max(eig["V_l"].shape[1], eig["V_r"].shape[1])
            if muse >= mevecs:
                raise ValueError(
                    f"Maximum specified eigenvector `{muse}` is larger than the number of eigenvectors `{mevecs}`. "
                    f"Use `.compute_eigendecomposition(k={muse})` to recompute the eigendecomposition."
                )

            return use

        eig = self.eigendecomposition
        if eig is None:
            raise RuntimeError("Compute eigendecomposition first as `.compute_eigendecomposition()`.")
        use = convert_use(use)

        start = logg.info("Computing terminal states")
        # we check for complex values only in the left, that's okay because the complex pattern
        # will be identical for left and right
        V_l, V_r = eig["V_l"][:, use], eig["V_r"].real[:, use]
        V_l = _complex_warning(V_l, use, use_imag=False)

        # retrieve embedding and concatenate
        if basis is not None:
            X_em = _get_basis(self.adata, basis=basis)[:, :n_comps]
            X = np.concatenate([V_r, X_em], axis=1)
        else:
            X = V_r

        # filter out cells which are in the lowest q percentile in abs value in each eigenvector
        if percentile is not None:
            logg.debug("Filtering out cells according to percentile")
            if not (0 <= percentile <= 100):
                raise ValueError(f"Expected `percentile` to be in interval `[0, 100]`, found `{percentile}`.")
            cutoffs = np.percentile(np.abs(V_l), percentile, axis=0)
            ixs = np.any(cutoffs <= np.abs(V_l), axis=1)
            X = X[ixs, :]

        if scale is None:
            scale = basis is not None
        if scale:
            X = st.zscore(X, axis=0)

        # cluster X
        if method == "kmeans" and n_clusters_kmeans is None:
            n_clusters_kmeans = len(use) + (percentile is None)
            if X.shape[0] < n_clusters_kmeans:
                raise ValueError(
                    f"Filtering resulted in only {X.shape[0]} cell(s), insufficient to cluster into "
                    f"`{n_clusters_kmeans}` clusters. Consider decreasing the value of `percentile`."
                )

        # fmt: off
        logg.debug(f"Using `{use}` eigenvectors, basis `{basis!r}` and method `{method!r}` for clustering")
        clusters = _cluster_X(
            X,
            method=method,
            n_clusters=n_clusters_kmeans,
            n_neighbors=n_neighbors,
            resolution=resolution,
        )

        # fill in the labels in case we filtered out cells before
        if percentile is not None:
            labels = np.repeat(None, len(self))
            labels[ixs] = clusters
        else:
            labels = clusters
        labels = pd.Series(labels, index=self.adata.obs_names, dtype="category")
        labels = labels.cat.rename_categories({c: str(c) for c in labels.cat.categories})

        # filtering to get rid of some of the leftover transient states
        if n_matches_min > 0:
            logg.debug(f"Filtering according to `n_matches_min={n_matches_min}`")
            distances = _get_connectivities(self.adata, mode="distances", n_neighbors=n_neighbors_filtering)
            labels = _filter_cells(distances, rc_labels=labels, n_matches_min=n_matches_min)
        # fmt: on

        return self.set_terminal_states(
            states=labels,
            cluster_key=cluster_key,
            probs=self._compute_term_states_probs(eig, use),
            params=self._create_params(),
            time=start,
        )

    def _compute_term_states_probs(self, eig: Dict[str, Any], use: List[int]) -> pd.Series:
        # get the truncated eigendecomposition
        V, evals = eig["V_l"].real[:, use], eig["D"].real[use]

        # shift and scale
        V_pos = np.abs(V)
        V_shifted = V_pos - np.min(V_pos, axis=0)
        V_scaled = V_shifted / np.max(V_shifted, axis=0)

        # check the ranges are correct
        np.testing.assert_allclose(np.min(V_scaled, axis=0), 0)
        np.testing.assert_allclose(np.max(V_scaled, axis=0), 1)

        # further scale by the eigenvalues
        V_eigs = V_scaled / evals

        # sum over cols and scale
        c = np.sum(V_eigs, axis=1)
        c /= np.max(c)

        return pd.Series(c, index=self.adata.obs_names)

    def _read_from_adata(self, adata: AnnData, **kwargs: Any) -> bool:
        ok = super()._read_from_adata(adata, **kwargs)
        ok = ok and self._read_eigendecomposition(adata, allow_missing=False)
        return ok and self._read_fate_probabilities(adata) and self._read_absorption_times(adata)



import abc
import types
from typing import Any, Dict, Literal, Optional, Sequence, Tuple, Union

import scvelo as scv

import numpy as np
import pandas as pd
import scipy.sparse as sp
from pandas.api.types import infer_dtype

from matplotlib.colors import to_hex

from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils._colors import (
    _convert_to_hex_colors,
    _create_categorical_colors,
    _map_names_and_colors,
)
from cellrank._utils._docs import d, inject_docs
from cellrank._utils._key import Key
from cellrank._utils._lineage import Lineage
from cellrank._utils._utils import (
    RandomKeys,
    _convert_to_categorical_series,
    _merge_categorical_series,
    _unique_order_preserving,
)
from cellrank.estimators._base_estimator import BaseEstimator
from cellrank.estimators.mixins._utils import (
    PlotMode,
    SafeGetter,
    StatesHolder,
    logger,
    shadow,
)
from cellrank.kernels._base_kernel import KernelExpression

__all__ = ["TermStatesEstimator"]


@d.dedent
class TermStatesEstimator(BaseEstimator, abc.ABC):
    """Base class for all estimators predicting the initial and terminal states.

    Parameters
    ----------
    %(base_estimator.parameters)s
    """

    def __init__(
        self,
        object: Union[AnnData, np.ndarray, sp.spmatrix, KernelExpression],
        **kwargs: Any,
    ):
        super().__init__(object=object, **kwargs)
        self._init_states = StatesHolder()
        self._term_states = StatesHolder()

    @property
    @d.get_summary(base="tse_term_states")
    def terminal_states(self) -> Optional[pd.Series]:
        """Categorical annotation of terminal states.

        By default, all transient cells will be labeled as `NaN`.
        """
        return self._term_states.assignment

    @property
    @d.get_summary(base="tse_term_states_probs")
    def terminal_states_probabilities(self) -> Optional[pd.Series]:
        """Probability to be a terminal state."""
        return self._term_states.probs

    @property
    @d.get_summary(base="tse_init_states")
    def initial_states(self) -> Optional[pd.Series]:
        """Categorical annotation of initial states.

        By default, all transient cells will be labeled as `NaN`.
        """
        return self._init_states.assignment

    @property
    @d.get_summary(base="tse_init_states_probs")
    def initial_states_probabilities(self) -> Optional[pd.Series]:
        """Probability to be an initial state."""
        return self._init_states.probs

    @d.dedent
    def set_terminal_states(
        self,
        states: Union[pd.Series, Dict[str, Sequence[Any]]],
        cluster_key: Optional[str] = None,
        allow_overlap: bool = False,
        **kwargs: Any,
    ) -> "TermStatesEstimator":
        """Set the :attr:`terminal_states`.

        Parameters
        ----------
        states
            States to select. Valid options are:

            - categorical :class:`~pandas.Series` where each category corresponds to an individual state.
              `NaN` entries denote cells that do not belong to any state, i.e., transient cells.
            - :class:`dict` where keys are states and values are lists of cell barcodes corresponding to
              annotations in :attr:`~anndata.AnnData.obs_names`.
              If only 1 key is provided, values should correspond to clusters if a categorical
              :class:`~pandas.Series` can be found in :attr:`~anndata.AnnData.obs`.
        cluster_key
            Key in :attr:`~anndata.AnnData.obs` to associate names and colors with :attr:`terminal_states`.
            Each state will be given the name and color corresponding to the cluster it mostly overlaps with.
        %(allow_overlap)s
        kwargs
            Additional keyword arguments.

        Returns
        -------
        Returns self and updates the following fields:

        - :attr:`terminal_states` - %(tse_term_states.summary)s
        - :attr:`terminal_states_probabilities` - %(tse_term_states_probs.summary)s
        """
        states, colors = self._set_categorical_labels(
            categories=states,
            cluster_key=cluster_key,
            existing=None,
        )
        self._write_states(
            "terminal",
            states=states,
            colors=colors,
            allow_overlap=allow_overlap,
            **kwargs,
        )
        return self

    @d.dedent
    def set_initial_states(
        self,
        states: Union[pd.Series, Dict[str, Sequence[Any]]],
        cluster_key: Optional[str] = None,
        allow_overlap: bool = False,
        **kwargs: Any,
    ) -> "TermStatesEstimator":
        """Set the :attr:`initial_states`.

        Parameters
        ----------
        states
            Which states to select. Valid options are:

            - categorical :class:`~pandas.Series` where each category corresponds to an individual state.
              `NaN` entries denote cells that do not belong to any state, i.e., transient cells.
            - :class:`dict` where keys are states and values are lists of cell barcodes corresponding to
              annotations in :attr:`~anndata.AnnData.obs_names`.
              If only 1 key is provided, values should correspond to clusters if a categorical
              :class:`~pandas.Series` can be found in :attr:`~anndata.AnnData.obs`.
        cluster_key
            Key in :attr:`~anndata.AnnData.obs` to associate names and colors :attr:`initial_states`.
            Each state will be given the name and color corresponding to the cluster it mostly overlaps with.
        %(allow_overlap)s
        kwargs
            Additional keyword arguments.

        Returns
        -------
        Returns self and updates the following fields:

        - :attr:`initial_states` - %(tse_init_states.summary)s
        - :attr:`initial_states_probabilities` - %(tse_init_states_probs.summary)s
        """
        states, colors = self._set_categorical_labels(
            categories=states,
            cluster_key=cluster_key,
            existing=None,
        )
        self._write_states(
            "initial",
            states=states,
            colors=colors,
            allow_overlap=allow_overlap,
            **kwargs,
        )
        return self

    @d.get_sections(base="tse_rename_term_states", sections=["Parameters", "Returns"])
    @d.get_full_description(base="tse_rename_term_states")
    @d.dedent
    def rename_terminal_states(self, old_new: Dict[str, str]) -> "TermStatesEstimator":
        """Rename the :attr:`terminal_states`.

        Parameters
        ----------
        old_new
            Dictionary that maps old names to unique new names.

        Returns
        -------
        Returns self and updates the following fields:

        - :attr:`terminal_states` - %(tse_term_states.summary)s
        """
        states = self.terminal_states
        if states is None:
            raise RuntimeError(
                "Compute terminal states first as `.predict_terminal_states()` or "
                "set them manually as `.set_terminal_states()`."
            )

        # fmt: off
        if not isinstance(old_new, dict):
            raise TypeError(f"Expected new names to be a `dict`, found `{type(old_new)}`.")
        if not len(old_new):
            return self

        old_names = states.cat.categories
        mask = np.isin(list(old_new.keys()), old_names)
        if not np.all(mask):
            invalid = sorted(np.array(list(old_new.keys()))[~mask])
            raise ValueError(f"Invalid terminal states names: `{invalid}`. Valid names are: `{sorted(old_names)}`.")

        names_after_renaming = [old_new.get(n, n) for n in old_names]
        if len(set(names_after_renaming)) != len(old_names):
            raise ValueError(f"After renaming, terminal states will no longer unique: `{names_after_renaming}`.")
        # fmt: on

        assignment = states.cat.rename_categories(old_new)
        memberships = self._term_states.memberships  # save before overwriting
        self._write_states(
            "terminal",
            states=assignment,
            colors=self._term_states.colors,
            probs=self.terminal_states_probabilities,
            log=False,
        )
        if memberships is not None:
            memberships.names = [old_new.get(n, n) for n in memberships.names]

        self._term_states = self._term_states.set(assignment=assignment, memberships=memberships)
        return self

    @d.dedent
    def rename_initial_states(self, old_new: Dict[str, str]) -> "TermStatesEstimator":
        """Rename the :attr:`initial_states`.

        Parameters
        ----------
        old_new
            Dictionary that maps old names to unique new names.

        Returns
        -------
        Returns self and updates the following fields:

        - :attr:`initial_states` - %(tse_init_states.summary)s
        """
        states = self.initial_states
        if states is None:
            raise RuntimeError(
                "Compute initial states first as `.predict_initial_states()` or "
                "set them manually as `.set_initial_states()`."
            )

        # fmt: off
        if not isinstance(old_new, dict):
            raise TypeError(f"Expected new names to be a `dict`, found `{type(old_new)}`.")
        if not len(old_new):
            return self

        old_names = states.cat.categories
        mask = np.isin(list(old_new.keys()), old_names)
        if not np.all(mask):
            invalid = sorted(np.array(list(old_new.keys()))[~mask])
            raise ValueError(f"Invalid terminal states names: `{invalid}`. Valid names are: `{sorted(old_names)}`.")

        names_after_renaming = [old_new.get(n, n) for n in old_names]
        if len(set(names_after_renaming)) != len(old_names):
            raise ValueError(f"After renaming, terminal states will no longer unique: `{names_after_renaming}`.")
        # fmt: on

        assignment = states.cat.rename_categories(old_new)
        memberships = self._init_states.memberships  # save overwriting
        self._write_states(
            "initial",
            states=assignment,
            colors=self._init_states.colors,
            probs=self.initial_states_probabilities,
            log=False,
        )
        if memberships is not None:
            memberships.names = [old_new.get(n, n) for n in memberships.names]

        self._init_states = self._init_states.set(assignment=assignment, memberships=memberships)
        return self

    @d.dedent
    @inject_docs(m=PlotMode)
    def plot_macrostates(
        self,
        which: Literal["all", "initial", "terminal"],
        states: Optional[Union[str, Sequence[str]]] = None,
        color: Optional[str] = None,
        discrete: bool = True,
        mode: Literal["embedding", "time"] = PlotMode.EMBEDDING,
        time_key: str = "latent_time",
        same_plot: bool = True,
        title: Optional[Union[str, Sequence[str]]] = None,
        cmap: str = "viridis",
        **kwargs: Any,
    ) -> None:
        """Plot macrostates on an embedding or along pseudotime.

        Parameters
        ----------
        which
            Which macrostates to plot. Valid options are:

            - ``'all'`` - plot all macrostates.
            - ``'initial'`` - plot macrostates marked as :attr:`initial_states`.
            - ``'terminal'`` - plot macrostates marked as :attr:`terminal_states`.
        states
            Subset of the macrostates to show. If :obj:`None`, plot all macrostates.
        color
            Key in :attr:`~anndata.AnnData.obs` or :attr:`~anndata.AnnData.var` used to color the observations.
        discrete
            Whether to plot the data as continuous or discrete observations.
            If the data cannot be plotted as continuous observations, it will be plotted as discrete.
        mode
            Whether to plot the probabilities in an embedding or along the pseudotime.
        time_key
            Key in :attr:`~anndata.AnnData.obs` where pseudotime is stored. Only used when ``mode = {m.TIME!r}``.
        title
            Title of the plot.
        same_plot
            Whether to plot the data on the same plot or not. Only use when ``mode = {m.EMBEDDING!r}``.
            If `True` and ``discrete = False``, ``color`` is ignored.
        cmap
            Colormap for continuous annotations.
        kwargs
            Keyword arguments for :func:`~scvelo.pl.scatter`.

        Returns
        -------
        %(just_plots)s
        """
        if which == "all":
            obj: Optional[StatesHolder] = getattr(self, "_macrostates", None)
            if obj is None:
                raise RuntimeError(f"`{type(self).__name__}` cannot plot macrostates.")
        elif which == "initial":
            obj = self._init_states
        elif which == "terminal":
            obj = self._term_states
        else:
            raise ValueError(
                f"Unable to plot `{which!r}` states. " f"Valid options are: `{['all', 'initial', 'terminal']}`."
            )

        name = "macrostates" if which == "macro" else f"{which} states"
        if obj.assignment is None and obj.memberships is None:
            raise RuntimeError(f"Compute {name} first.")

        if not discrete and obj.memberships is None:
            logg.warning(f"Unable to plot {name} in continuous mode, using discrete")
            discrete = True

        data = obj.assignment if discrete else obj.memberships
        colors = obj.colors

        if discrete:
            return self._plot_discrete(
                _data=data,
                _colors=colors,
                _title=name,
                states=states,
                color=color,
                same_plot=same_plot,
                title=title,
                cmap=cmap,
                **kwargs,
            )
        return self._plot_continuous(
            _data=data,
            _colors=colors,
            _title=name,
            states=states,
            color=color,
            mode=mode,
            time_key=time_key,
            same_plot=same_plot,
            title=title,
            cmap=cmap,
            **kwargs,
        )

    def _plot_discrete(
        self,
        _data: pd.Series,
        _colors: Optional[np.ndarray] = None,
        _title: Optional[str] = None,
        states: Optional[Union[str, Sequence[str]]] = None,
        color: Optional[str] = None,
        title: Optional[Union[str, Sequence[str]]] = None,
        same_plot: bool = True,
        cmap: str = "viridis",
        **kwargs: Any,
    ) -> None:
        if not isinstance(_data, pd.Series):
            raise TypeError(f"Expected `data` to be of type `pandas.Series`, found `{type(_data)}`.")
        if not isinstance(_data.dtype, pd.CategoricalDtype):
            raise TypeError(f"Expected `data` to be `categorical`, found `{infer_dtype(_data)}`.")

        names = list(_data.cat.categories)
        if _colors is None:
            _colors = _create_categorical_colors(len(names))
        if len(_colors) != len(names):
            raise ValueError(f"Expected `colors` to be of length `{len(names)}`, found `{len(_colors)}`.")
        color_mapper = dict(zip(names, _colors))

        states = _unique_order_preserving(states or names)
        if not len(states):
            raise ValueError("No states have been selected.")

        for name in states:
            if name not in names:
                raise ValueError(f"Invalid name `{name!r}`. Valid options are: `{sorted(names)}`.")
        _data = _data.cat.set_categories(states)

        color = [] if color is None else (color,) if isinstance(color, str) else color
        color = _unique_order_preserving(color)

        same_plot = same_plot or len(names) == 1
        kwargs.setdefault("legend_loc", "on data")
        kwargs["color_map"] = cmap

        # fmt: off
        with RandomKeys(self.adata, n=1 if same_plot else len(states), where="obs") as keys:
            if same_plot:
                outline = _data.cat.categories.to_list()
                _data = _data.cat.add_categories(["nan"]).fillna("nan")
                states.append("nan")
                color_mapper["nan"] = "#dedede"
                self.adata.obs[keys[0]] = _data
                self.adata.uns[f"{keys[0]}_colors"] = [color_mapper[name] for name in states]
                title = _title if title is None else title
            else:
                outline = None
                for key, cat in zip(keys, states):
                    self.adata.obs[key] = _data.cat.set_categories([cat])
                    self.adata.uns[f"{key}_colors"] = [color_mapper[cat]]
                title = [f"{_title} {name}" for name in states] if title is None else title

            if isinstance(title, str):
                title = [title]

            scv.pl.scatter(
                self.adata,
                color=color + keys,
                title=color + title,
                add_outline=outline,
                **kwargs,
            )
        # fmt: on

    def _plot_continuous(
        self,
        _data: Lineage,
        _colors: Optional[np.ndarray] = None,
        _title: Optional[str] = None,
        states: Optional[Union[str, Sequence[str]]] = None,
        color: Optional[str] = None,
        mode: Literal["embedding", "time"] = PlotMode.EMBEDDING,
        time_key: str = "latent_time",
        title: Optional[Union[str, Sequence[str]]] = None,
        same_plot: bool = True,
        cmap: str = "viridis",
        **kwargs: Any,
    ) -> None:
        mode = PlotMode(mode)
        if not isinstance(_data, Lineage):
            raise TypeError(f"Expected data to be of type `Lineage`, found `{type(_data)}`.")

        if states is None:
            states = _data.names
        if not len(states):
            raise ValueError("No lineages have been selected.")
        is_singleton = _data.shape[1] == 1
        _data = _data[states].copy()

        if mode == "time" and same_plot:
            logg.warning("Invalid combination `mode='time'` and `same_plot=True`. Using `same_plot=False`")
            same_plot = False

        _data_X = _data.X  # list(_data.T) behaves differently than a numpy.array
        if _data_X.shape[1] == 1:
            same_plot = False
            if np.allclose(_data_X, 1.0):
                # matplotlib shows even tiny perturbations in the colormap
                _data_X = np.ones_like(_data_X)

        for col in _data_X.T:
            mask = ~np.isclose(col, 1.0)
            # change the maximum value - the 1 is artificial and obscures the color scaling
            if np.any(mask):
                col[~mask] = np.nanmax(col[mask])

        # fmt: off
        color = [] if color is None else (color,) if isinstance(color, str) else color
        color = _unique_order_preserving(color)

        if mode == PlotMode.TIME:
            kwargs.setdefault("legend_loc", "best")
            if time_key is None:
                raise KeyError(
                    "The name of the column in `adata.obs` defining the pseudotime needs to be defined via the "
                    "`time_key` argument."
                )
            if title is None:
                title = [f"{_title} {state}" for state in states]
            if time_key not in self.adata.obs:
                raise KeyError(f"Unable to find pseudotime in `adata.obs[{time_key!r}]`.")
            if len(color) and len(color) not in (1, _data_X.shape[1]):
                raise ValueError(f"Expected `color` to be of length `1` or `{_data_X.shape[1]}`, "
                                 f"found `{len(color)}`.")
            kwargs["x"] = self.adata.obs[time_key]
            kwargs["y"] = list(_data_X.T)
            kwargs["color"] = color if len(color) else None
            kwargs["xlabel"] = [time_key] * len(states)
            kwargs["ylabel"] = ["probability"] * len(states)
        elif mode == PlotMode.EMBEDDING:
            kwargs.setdefault("legend_loc", "on data")
            if same_plot:
                if color:
                    # https://github.com/theislab/scvelo/issues/673
                    logg.warning("Ignoring `color` when `mode='embedding'` and `same_plot=True`")
                title = [_title] if title is None else title
                kwargs["color_gradients"] = _data
            else:
                title = [f"{_title} {state}" for state in states] if title is None else title
                if isinstance(title, str):
                    title = [title]
                title = color + title
                kwargs["color"] = color + list(_data_X.T)
        else:
            raise NotImplementedError(f"Mode `{mode}` is not yet implemented.")
        # fmt: on

        # e.g. a stationary distribution
        if is_singleton and not np.allclose(_data_X, 1.0):
            kwargs.setdefault("perc", [0, 95])
            _ = kwargs.pop("color_gradients", None)

        scv.pl.scatter(
            self.adata,
            title=title,
            color_map=cmap,
            **kwargs,
        )

    def _set_categorical_labels(
        self,
        categories: Union[pd.Series, Dict[str, Any]],
        cluster_key: Optional[str] = None,
        existing: Optional[pd.Series] = None,
    ) -> Tuple[pd.Series, np.ndarray]:
        # fmt: off
        if isinstance(categories, dict):
            key = next(iter(categories.keys()))
            data = self.adata.obs.get(key, None)
            is_categorical = data is not None and isinstance(data.dtype, pd.CategoricalDtype)
            if len(categories) == 1 and is_categorical:
                vals = categories[key]
                if isinstance(vals, str) or not isinstance(vals, Sequence):
                    vals = (categories[key],)

                clusters = self.adata.obs[key]
                clusters = clusters.cat.rename_categories({c: str(c) for c in clusters.cat.categories})
                vals = tuple(str(v) for v in vals)
                categories = {cat: self.adata[clusters == cat].obs_names for cat in vals}

            categories = _convert_to_categorical_series(categories, list(self.adata.obs_names))
        if not isinstance(categories.dtype, pd.CategoricalDtype):
            raise TypeError(f"Expected object to be `categorical`, found `{infer_dtype(categories)}`.")

        if existing is not None:
            categories = _merge_categorical_series(old=existing, new=categories)

        if cluster_key is not None:
            # check that we can load the reference series from adata
            if cluster_key not in self.adata.obs:
                raise KeyError(f"Unable to find clusters in `adata.obs[{cluster_key!r}]`.")
            series_query, series_reference = categories, self.adata.obs[cluster_key]
            series_reference = series_reference.cat.rename_categories(
                {c: str(c) for c in series_reference.cat.categories}
            )

            # load the reference colors if they exist
            if Key.uns.colors(cluster_key) in self.adata.uns:
                colors_reference = _convert_to_hex_colors(self.adata.uns[Key.uns.colors(cluster_key)])
            else:
                colors_reference = _create_categorical_colors(len(series_reference.cat.categories))

            names, colors = _map_names_and_colors(
                series_reference=series_reference,
                series_query=series_query,
                colors_reference=colors_reference,
            )
            cats = categories.cat.categories
            categories = categories.cat.rename_categories(dict(zip(cats, names)))
        else:
            colors = _create_categorical_colors(len(categories.cat.categories))

        return categories, colors
        # fmt: on

    @logger
    @shadow
    def _write_states(
        self,
        which: Literal["initial", "terminal"],
        states: Optional[pd.Series],
        colors: Optional[np.ndarray],
        probs: Optional[pd.Series] = None,
        params: Dict[str, Any] = types.MappingProxyType({}),
        allow_overlap: bool = False,
    ) -> str:
        # fmt: off
        backward = which == "initial"
        if not allow_overlap:
            fwd_states, bwd_states = (self.terminal_states, states) if backward else (states, self.initial_states)
            if fwd_states is not None and bwd_states is not None:
                overlap = np.sum(~pd.isnull(fwd_states) & ~pd.isnull(bwd_states))
                if overlap:
                    raise ValueError(
                        f"Found `{overlap}` overlapping cells between initial and terminal states. "
                        f"If this is intended, please use `allow_overlap=True`."
                    )

        key = Key.obs.term_states(self.backward, bwd=backward)
        self._set(obj=self.adata.obs, key=key, value=states)
        self._set(obj=self.adata.obs, key=Key.obs.probs(key), value=probs)
        self._set(obj=self.adata.uns, key=Key.uns.colors(key), value=colors)
        if backward:
            self._init_states = self._init_states.set(assignment=states, probs=probs, colors=colors)
        else:
            self._term_states = self._term_states.set(assignment=states, probs=probs, colors=colors)
        self.params[key] = dict(params)
        # fmt: on

        return (
            f"Adding `adata.obs[{key!r}]`\n"
            f"       `adata.obs[{Key.obs.probs(key)!r}]`\n"
            f"       `.{which}_states`\n"
            f"       `.{which}_states_probabilities`\n"
            "    Finish"
        )

    def _read_from_adata(self, adata: AnnData, **kwargs: Any) -> bool:
        ok = super()._read_from_adata(adata, **kwargs)
        if not ok:
            return False

        # fmt: off
        for backward in [True, False]:
            key = Key.obs.term_states(self.backward, bwd=backward)
            with SafeGetter(self, allowed=KeyError) as sg:
                assignment = self._get(obj=adata.obs, key=key, shadow_attr="obs", dtype=pd.Series)
                probs = self._get(obj=adata.obs, key=Key.obs.probs(key), shadow_attr="obs", dtype=pd.Series)
                colors = self._get(obj=adata.uns, key=Key.uns.colors(key), shadow_attr="uns",
                                   dtype=(list, tuple, np.ndarray))
                colors = np.asarray([to_hex(c) for c in colors])
                if backward:
                    self._init_states = StatesHolder(assignment=assignment, probs=probs, colors=colors)
                else:
                    self._term_states = StatesHolder(assignment=assignment, probs=probs, colors=colors)
                self.params[key] = self._read_params(key)
        # fmt: on

        # status is based on `backward=False` by design
        return sg.ok

    def _format_params(self) -> str:
        fmt = super()._format_params()
        # fmt: off
        init_states = None if self.initial_states is None else sorted(self.initial_states.cat.categories)
        term_states = None if self.terminal_states is None else sorted(self.terminal_states.cat.categories)
        return fmt + f", initial_states={init_states}, terminal_states={term_states}"
        # fmt: on



import copy as copy_
import enum
import functools
import inspect
import itertools
import pathlib
import types
from typing import (
    Any,
    Callable,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import numpy as np
import pandas as pd
import scipy.stats as st
from pandas.api.types import infer_dtype

import matplotlib.pyplot as plt
from matplotlib import colors

from anndata import AnnData
from anndata._io.specs.methods import H5Group, ZarrGroup, write_basic
from anndata._io.specs.registry import _REGISTRY, IOSpec

from cellrank import logging as logg
from cellrank._utils._colors import (
    _compute_mean_color,
    _create_categorical_colors,
    _get_bg_fg_colors,
)
from cellrank._utils._docs import d, inject_docs
from cellrank._utils._enum import ModeEnum
from cellrank._utils._key import Key

__all__ = ["Lineage", "LineageView"]

ColorLike = TypeVar("ColorLike")
_ERROR_NOT_ITERABLE = "Expected `{}` to be iterable, found type `{}`."
_ERROR_WRONG_SIZE = "Expected `{}` to be of size `{{}}`, found `{{}}`."

_HT_CELLS = 10  # head and tail cells to show
_HTML_REPR_THRESH = 100
_DUMMY_CELL = "<td style='text-align: right;'>...</td>"
_ORDER = "C"


class PrimingDegree(ModeEnum):  # noqa: D101
    KL_DIVERGENCE = enum.auto()
    ENTROPY = enum.auto()


class DistanceMeasure(ModeEnum):  # noqa: D101
    COSINE_SIM = enum.auto()
    WASSERSTEIN_DIST = enum.auto()
    KL_DIV = enum.auto()
    JS_DIV = enum.auto()
    MUTUAL_INFO = enum.auto()
    EQUAL = enum.auto()


class NormWeights(ModeEnum):  # noqa: D101
    SCALE = enum.auto()
    SOFTMAX = enum.auto()


class Reduction(ModeEnum):  # noqa: D101
    DIST = enum.auto()
    SCALE = enum.auto()


class LinKind(ModeEnum):  # noqa: D101
    MACROSTATES = enum.auto()
    TERM_STATES = enum.auto()
    FATE_PROBS = enum.auto()


def _at_least_2d(array: np.ndarray, dim: int):
    return np.expand_dims(array, dim) if array.ndim < 2 else array


def wrap(numpy_func: Callable) -> Callable:
    """Wrap a :mod:`numpy` function.

    Modifies functionality of some function (e.g. ignoring `.squeeze`, retaining dimensions).

    Parameters
    ----------
    numpy_func
        Function to be wrapped.

    Returns
    -------
    Wrapped function which takes a :class:`cellrank.Lineage` and return :class:`cellrank.Lineage`.
    """

    @functools.wraps(numpy_func)
    def decorator(array: "Lineage", *args, **kwargs):
        if not isinstance(array, Lineage):
            raise TypeError(f"Expected array to be of type `Lineage`, found `{type(array).__name__}`.")
        if fname == "squeeze":
            return array
        if fname == "array_repr":
            return repr(array)

        if "axis" in kwargs:
            axis = kwargs["axis"]
        elif axis_ix < len(args):
            axis = args[axis_ix]
        else:
            axis = default_axis

        res = np.array(numpy_func(array.X, *args, **kwargs), copy=False)

        # handle expand_dim
        if res.ndim > 2:
            return array

        # handle reductions
        if not res.shape:
            return Lineage(np.array([[res]]), names=[fname], colors=["grey"])
        if res.shape == array.shape:
            return Lineage(res, names=array.names, colors=array.colors)

        res = np.expand_dims(res, axis)

        is_t = int(array._is_transposed)
        if is_t:
            res = res.T

        lin = None
        if res.shape[0] == array.shape[is_t]:
            lin = Lineage(
                res,
                names=[f"{fname} of {', '.join(array.names)}"],
                colors=["grey"],
            )

        if res.shape[1] == array.shape[1 - is_t]:
            lin = Lineage(res, names=[f"{fname} of {n}" for n in array.names], colors=array.colors)

        if lin is not None:
            return lin.T if is_t else lin

        raise RuntimeError(
            f"Unable to interpret result of function `{fname}` called " f"with args `{args}`, kwargs: `{kwargs}`."
        )

    params = inspect.signature(numpy_func).parameters
    if "axis" in params:
        axis_ix = list(params.keys()).index("axis") - 1
        default_axis = params["axis"].default
    else:
        axis_ix = 256
        default_axis = None
    assert axis_ix >= 0, f"Expected argument `'axis'` not to be first for function `{numpy_func.__name__}`."

    fname = numpy_func.__name__
    if fname == "amin":
        fname = "min"
    elif fname == "amax":
        fname = "max"

    return decorator


def _register_handled_functions():
    # adapted from:
    # https://github.com/numpy/numpy/blob/v1.26.0/numpy/testing/overrides.py#L50
    try:
        from numpy.core.overrides import ARRAY_FUNCTIONS
    except ImportError:
        ARRAY_FUNCTIONS = [getattr(np, attr) for attr in dir(np)]

    handled_fns = {}
    for fn in ARRAY_FUNCTIONS:
        try:
            sig = inspect.signature(fn)
            if "axis" in sig.parameters:
                handled_fns[fn] = wrap(fn)
        except Exception:  # noqa: BLE001
            pass

    handled_fns.pop(np.expand_dims, None)

    handled_fns[np.allclose] = wrap(np.allclose)
    handled_fns[np.array_repr] = wrap(np.array_repr)
    handled_fns[st.entropy] = wrap(st.entropy)  # qol change

    return handled_fns


_HANDLED_FUNCTIONS = _register_handled_functions()


class LineageMeta(type):
    """Metaclass for Lineage.

    It registers functions which are handled by us and overloads common attributes, such as `.sum` with these functions.
    """

    __overloaded_functions__ = dict(  # noqa
        sum=np.sum,
        mean=np.mean,
        min=np.min,
        argmin=np.argmin,
        max=np.max,
        argmax=np.argmax,
        std=np.std,
        var=np.var,
        sort=np.sort,
        squeeze=np.squeeze,
        entropy=st.entropy,
    )

    def __new__(cls, clsname, superclasses, attributedict):  # noqa
        res = type.__new__(cls, clsname, superclasses, attributedict)
        for attrname, fn in LineageMeta.__overloaded_functions__.items():
            wrapped_fn = _HANDLED_FUNCTIONS.get(fn, None)
            if wrapped_fn:
                setattr(res, attrname, wrapped_fn)

        return res


class Lineage(np.ndarray, metaclass=LineageMeta):
    """Lightweight :class:`~numpy.ndarray` wrapper that adds names and colors.

    Parameters
    ----------
    input_array
        Input array containing lineage probabilities stored in columns.
    names
        Lineage names.
    colors
        Lineage colors.
    """

    def __new__(
        cls,
        input_array: np.ndarray,
        *,
        names: Iterable[str],
        colors: Optional[Iterable[ColorLike]] = None,
    ) -> "Lineage":
        """Create and return a new object."""
        if not isinstance(input_array, np.ndarray):
            raise TypeError(f"Input array must be of type `numpy.ndarray`, found `{type(input_array).__name__!r}`.")

        if input_array.ndim == 1:
            input_array = np.expand_dims(input_array, -1)
        elif input_array.ndim > 2:
            raise ValueError(f"Input array must be 2-dimensional, found `{input_array.ndim}`.")

        if not input_array.shape[0]:
            raise ValueError("Expected number cells to be at least `1`, found `0`.")
        if not input_array.shape[1]:
            raise ValueError("Expected number of lineages to be at least `1`, found `0`.")

        obj = np.array(input_array, copy=True).view(cls)
        obj._n_lineages = obj.shape[1]
        obj._is_transposed = False
        obj.names = names  # these always create a copy, which is a good thing
        obj.colors = colors

        return obj

    def __array_finalize__(self, obj) -> None:
        if obj is None:
            return

        _names = getattr(obj, "_names", None)
        if _names is not None:
            self._names = _names
            self._n_lineages = len(_names)
            self._names_to_ixs = {n: i for i, n in enumerate(self.names)}
        else:
            self._names = None
            self._names_to_ixs = None
            self._n_lineages = getattr(obj, "_n_lineages", obj.shape[1] if obj.ndim == 2 else 0)

        self._colors = getattr(obj, "colors", None)
        self._is_transposed = getattr(obj, "_is_transposed", False)

    def __array_function__(self, func, types, args, kwargs):
        if func not in _HANDLED_FUNCTIONS:
            return NotImplemented
        # Note: this allows subclasses that don't override
        # __array_function__ to handle MyArray objects
        if not all(issubclass(t, self.__class__) for t in types):
            return NotImplemented

        return _HANDLED_FUNCTIONS[func](*args, **kwargs)

    def __getitem__(self, item) -> "Lineage":
        was_transposed = False
        if self._is_transposed:
            was_transposed = True
            self = self.T
            if isinstance(item, tuple):
                item = item[::-1]

        obj = self.__getitem(item)

        return obj.T if was_transposed else obj

    def _mix_lineages(self, rows, mixtures: Iterable[Union[str, Any]]) -> "Lineage":
        from cellrank._utils._utils import _unique_order_preserving

        def unsplit(names: str) -> Tuple[str, ...]:
            return tuple(sorted({name.strip(" ") for name in names.strip(" ,").split(",")}))

        keys = [
            tuple(self._maybe_convert_names(unsplit(mixture), default=mixture))
            if isinstance(mixture, str)
            else (mixture,)
            for mixture in mixtures
        ]
        keys = _unique_order_preserving(keys)

        # check the `keys` are unique
        overlap = [set(ks) for ks in keys]
        for c1, c2 in itertools.combinations(overlap, 2):
            overlap = c1 & c2
            if overlap:
                raise ValueError(f"Found overlapping keys: `{self.names[list(overlap)]}`.")

        names, colors, res = [], [], []
        for key in map(list, keys):
            if key:
                res.append(self[rows, key].X.sum(1))
                names.append(", ".join(self.names[key]))
                colors.append(_compute_mean_color(self.colors[key]))

        return Lineage(np.stack(res, axis=-1), names=names, colors=colors)

    def __getitem(self, item):
        if isinstance(item, tuple):
            if len(item) > 2:
                raise ValueError(f"Expected key to be of length `2`, found `{len(item)}`.")

            item = list(item)
            if item[0] is Ellipsis or item[0] is None:
                item[0] = range(self.shape[0])
            if len(item) == 2 and (item[1] is Ellipsis or item[1] is None):
                item[1] = range(self.shape[1])
            item = tuple(item)

        is_tuple_len_2 = (
            isinstance(item, tuple)
            and len(item) == 2
            and isinstance(item[0], (int, np.integer, range, slice, tuple, list, np.ndarray))
        )
        if is_tuple_len_2:
            rows, col = item
            if isinstance(col, (int, np.integer, str)):
                col = [col]
            try:
                # slicing an array where row/col are like 2D indices
                if 1 < len(col) == len(rows) and len(rows) == self.shape[0]:
                    col = self._maybe_convert_names(col, make_unique=False)
                    return Lineage(
                        # never remove this expand_dims - it's critical
                        np.expand_dims(self.X[rows, col], axis=-1),
                        names=["mixture"],
                        colors=["#000000"],
                    )
            except TypeError:  # because of range
                pass

            if isinstance(col, (list, tuple, np.ndarray)):
                if any(isinstance(i, str) and "," in i for i in col):
                    return self._mix_lineages(rows, col)
                col = self._maybe_convert_names(col)
                item = rows, col
        else:
            if isinstance(item, (int, np.integer, str)):
                item = [item]
            col = range(len(self.names))
            if isinstance(item, (tuple, list, np.ndarray)):
                if any(isinstance(i, str) and "," in i for i in item):
                    return self._mix_lineages(slice(None, None, None), item)
                if any(isinstance(i, str) for i in item):
                    item = (slice(None, None, None), self._maybe_convert_names(item))
                    col = item[1]

        shape, row_order, col_order = None, None, None
        if is_tuple_len_2 and not isinstance(item[0], slice) and not isinstance(item[1], slice):
            item_0 = np.array(item[0]) if not isinstance(item[0], np.ndarray) else item[0]
            item_1 = np.array(item[1]) if not isinstance(item[1], np.ndarray) else item[1]
            item_0 = _at_least_2d(item_0, -1)
            item_1 = _at_least_2d(item_1, 0)

            # handle boolean indexing
            if item_1.dtype == bool:
                if item_0.dtype != bool:
                    if not issubclass(item_0.dtype.type, np.integer):
                        raise TypeError(f"Invalid type `{item_0.dtype.type}`.")
                    row_order = (
                        item_0[:, 0] if item_0.shape[0] == self.shape[0] else np.argsort(np.argsort(item_0[:, 0]))
                    )
                    item_0 = _at_least_2d(np.isin(np.arange(self.shape[0]), item_0), -1)
                item = item_0 * item_1
                shape = np.max(np.sum(item, axis=0)), np.max(np.sum(item, axis=1))
                col = np.where(np.all(item_1, axis=0))[0]
            elif item_0.dtype == bool:
                if item_1.dtype != bool:
                    if not issubclass(item_1.dtype.type, np.integer):
                        raise TypeError(f"Invalid type `{item_1.dtype.type}`.")
                    col_order = (
                        item_1[0, :] if item_1.shape[1] == self.shape[1] else np.argsort(np.argsort(item_1[0, :]))
                    )
                    item_1 = _at_least_2d(np.isin(np.arange(self.shape[1]), item_1), 0)

                item = item_0 * item_1
                shape = np.max(np.sum(item, axis=0)), np.max(np.sum(item, axis=1))
                col = np.where(np.all(item_1, axis=0))[0]
            else:
                # defer to numpy
                item = (item_0, item_1)

        obj = super().__getitem__(item)

        # keep the resulting shape
        if shape is not None:
            obj = obj.reshape(shape)
        # correctly reorder
        if row_order is not None:
            obj = obj[row_order, :]
        if col_order is not None:
            obj = obj[:, col_order]

        # correctly set names and colors
        if isinstance(obj, Lineage):
            obj._names = np.atleast_1d(self.names[col])
            obj._colors = np.atleast_1d(self.colors[col])
            if col_order is not None:
                obj._names = obj._names[col_order]
                obj._colors = obj._colors[col_order]
            obj._names_to_ixs = {name: i for i, name in enumerate(obj._names)}

        return obj

    @property
    def names(self) -> np.ndarray:
        """Lineage names."""
        return self._names

    @names.setter
    def names(self, value: Iterable[str]) -> None:
        if not isinstance(value, Iterable):
            raise TypeError(_ERROR_NOT_ITERABLE.format("names", type(value).__name__))

        value = [str(v) for v in value]
        value = self._check_axis1_shape(value, _ERROR_WRONG_SIZE.format("names"))

        if len(set(value)) != len(value):
            raise ValueError(f"Not all lineage names are unique: `{value}`.")

        self._names = self._prepare_annotation(value)
        self._names_to_ixs = {name: ix for ix, name in enumerate(self.names)}

    @property
    def colors(self) -> np.ndarray:
        """Lineage colors."""
        return self._colors

    @colors.setter
    def colors(self, value: Optional[Iterable[ColorLike]]) -> None:
        if value is None:
            value = _create_categorical_colors(self._n_lineages)
        elif not isinstance(value, Iterable):
            raise TypeError(_ERROR_NOT_ITERABLE.format("colors", type(value).__name__))

        value = self._check_axis1_shape(value, _ERROR_WRONG_SIZE.format("colors"))
        self._colors = self._prepare_annotation(
            value,
            checker=colors.is_color_like,
            transformer=colors.to_hex,
            checker_msg="Value `{}` is not a valid color.",
        )

    @property
    def X(self) -> np.ndarray:
        """Convert self to an array."""
        return np.array(self, copy=False)

    @property
    def T(self):
        """Transpose of self."""
        obj = self.transpose()
        obj._is_transposed = not self._is_transposed
        return obj

    @property
    def nlin(self) -> int:
        """Number of lineages."""
        return self.shape[1]

    @d.get_full_description(base="lin_pd")
    @d.get_sections(base="lin_pd", sections=["Parameters", "Returns"])
    def priming_degree(
        self,
        method: Literal["kl_divergence", "entropy"] = "kl_divergence",
        early_cells: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Compute the degree of lineage priming.

        It returns a score in :math:`[0, 1]` where :math:`0` stands for naive and :math:`1` stands for committed.

        Parameters
        ----------
        method
            The method used to compute the degree of lineage priming. Valid options are:

            - ``'kl_divergence'`` - as in :cite:`velten:17`, computes KL-divergence between the fate probabilities of
              a cell and the average fate probabilities. Computation of average fate probabilities can be restricted
              to a set of user-defined ``early_cells``.
            - ``'entropy'`` - as in :cite:`setty:19`, computes entropy over a cell's fate probabilities.
        early_cells
            Cell IDs or a mask marking early cells. If :obj:`None`, use all cells.
            Only used when ``method = 'kl_divergence'``.

        Returns
        -------
        The priming degree.
        """
        early_cells = np.ones((len(self),), dtype=np.bool_) if early_cells is None else np.asarray(early_cells)
        if not np.issubdtype(early_cells.dtype, np.bool_):
            early_cells = np.unique(early_cells)

        method = PrimingDegree(method)
        probs = self.X
        early_subset = probs[early_cells, :]
        if not len(early_subset):
            raise ValueError("No early cells have been specified.")

        with np.errstate(divide="ignore", invalid="ignore"):
            if method == PrimingDegree.KL_DIVERGENCE:
                probs = np.nan_to_num(
                    np.sum(probs * np.log2(probs / np.mean(early_subset, axis=0)), axis=1),
                    nan=1.0,
                    copy=False,
                )
            elif method == PrimingDegree.ENTROPY:
                probs = st.entropy(probs, axis=1)
                probs = np.max(probs) - probs
            else:
                raise NotImplementedError(f"Method `{method}` is not yet implemented")

            minn, maxx = np.min(probs), np.max(probs)
            return (probs - minn) / (maxx - minn)

    @d.dedent
    def plot_pie(
        self,
        reduction: Callable,
        title: Optional[str] = None,
        legend_loc: Optional[str] = "on data",
        legend_kwargs: Mapping = types.MappingProxyType({}),
        figsize: Optional[Tuple[float, float]] = None,
        dpi: Optional[float] = None,
        save: Optional[Union[pathlib.Path, str]] = None,
        **kwargs: Any,
    ) -> None:
        """Plot a pie chart visualizing aggregated lineage probabilities.

        Parameters
        ----------
        reduction
            Function that will be applied lineage-wise.
        title
            Title of the figure.
        legend_loc
            Location of the legend. If :obj:`None`, it is not shown.
        legend_kwargs
            Keyword arguments for :meth:`~matplotlib.axes.Axes.legend`.
        %(plotting)s
        kwargs
            Keyword arguments for :meth:`~matplotlib.axes.Axes.pie`.

        Returns
        -------
        %(just_plots)s
        """
        from cellrank._utils._utils import save_fig

        if len(self.names) == 1:
            raise ValueError("Cannot plot pie chart for only 1 lineage.")

        fig, ax = plt.subplots(figsize=figsize, dpi=dpi)

        if "autopct" not in kwargs:
            autopct_found = False
            autopct = "{:.1f}%".format  # we don't really care, we don't shot the pct, but the value
        else:
            autopct_found = True
            autopct = kwargs.pop("autopct")

        if title is None:
            title = reduction.__name__ if hasattr(reduction, "__name__") else None

        reduction = reduction(self, axis=int(self._is_transposed)).X.squeeze()
        reduction_norm = reduction / np.sum(reduction)

        wedges, texts, *autotexts = ax.pie(
            reduction_norm.squeeze(),
            labels=self.names if legend_loc == "on data" else None,
            autopct=autopct,
            wedgeprops={"edgecolor": "w"},
            colors=self.colors,
            **kwargs,
        )

        # if autopct is not None
        if len(autotexts):
            autotexts = autotexts[0]
            for name, at in zip(self.names, autotexts):
                ix = self._names_to_ixs[name]
                at.set_color(_get_bg_fg_colors(self.colors[ix])[1])
                if not autopct_found:
                    at.set_text(f"{reduction[ix]:.4f}")

        if legend_loc not in (None, "none", "on data"):
            ax.legend(
                wedges,
                self.names,
                title="lineages",
                loc=legend_loc,
                **legend_kwargs,
            )

        ax.set_title(title)
        ax.set_aspect("equal")

        if save is not None:
            save_fig(fig, save)

    @d.dedent
    @inject_docs(m=Reduction, dm=DistanceMeasure, nw=NormWeights)
    def reduce(
        self,
        *keys: str,
        mode: Literal["dist", "scale"] = Reduction.DIST,
        dist_measure: Literal[
            "cosine_sim", "wasserstein_dist", "kl_div", "js_div", "mutual_info", "equal"
        ] = DistanceMeasure.MUTUAL_INFO,
        normalize_weights: Literal["scale", "softmax"] = NormWeights.SOFTMAX,
        softmax_scale: float = 1.0,
        return_weights: bool = False,
    ) -> Union["Lineage", Tuple["Lineage", Optional[pd.DataFrame]]]:
        """Subset states and normalize them so that they again sum to :math:`1`.

        Parameters
        ----------
        keys
            List of keys that define the states, to which this object will be reduced by projecting the values
            of the other states.
        mode
            Reduction mode to use. Valid options are:

            - ``{m.DIST!r}`` - use a distance measure ``dist_measure`` to compute weights.
            - ``{m.SCALE!r}`` - just rescale the values.
        dist_measure
            Used to quantify similarity between query and reference states. Valid options are:

            - ``{dm.COSINE_SIM!r}`` - cosine similarity.
            - ``{dm.WASSERSTEIN_DIST!r}`` - Wasserstein distance.
            - ``{dm.KL_DIV!r}`` - Kullback–Leibler divergence.
            - ``{dm.JS_DIV!r}`` - Jensen–Shannon divergence.
            - ``{dm.MUTUAL_INFO!r}`` - mutual information.
            - ``{dm.EQUAL!r}`` - equally redistribute the mass among the rest.

            Only use when ``mode = {m.DIST!r}``.
        normalize_weights
            How to row-normalize the weights. Valid options are:

            - ``{nw.SCALE!r}`` - divide by the sum.
            - ``{nw.SOFTMAX!r}``- use a softmax.

            Only used when ``mode = {m.DIST!r}``.
        softmax_scale
            Scaling factor in the softmax, used for normalizing the weights to sum to :math:`1`.
        return_weights
            If `True`, a :class:`~pandas.DataFrame` of the weights used for the projection is also returned.
            If ``mode = {m.SCALE!r}``, the weights will be `None`.

        Returns
        -------
        The lineage object, reduced to the %(initial_or_terminal)s states.
        The weights used for the projection of shape ``(n_query, n_reference)``, if ``return_weights = True``.
        """
        mode = Reduction(mode)
        dist_measure = DistanceMeasure(dist_measure)
        normalize_weights = NormWeights(normalize_weights)

        if self._is_transposed:
            raise RuntimeError("This method works only on non-transposed lineages.")

        if not len(keys):
            raise ValueError("Unable to perform the reduction, no keys specified.")

        # check the lineage object
        if not np.allclose(np.sum(self.X, axis=1), 1.0):
            raise ValueError("Memberships do not sum to one row-wise.")

        if len(keys) == 1:
            tmp = self[:, keys]
            return Lineage(
                np.ones((self.shape[0], 1), dtype=self.dtype),
                names=tmp.names,
                colors=tmp.colors,
            )

        # check input parameters
        if return_weights and mode == Reduction.SCALE:
            logg.warning(f"If `mode={mode!r}`, no weights are computed. Returning `None`")

        reference = self[:, keys]
        rest = [k for k in self.names if all(k not in rk for rk in reference.names)]
        if not rest:
            logg.warning(
                "Unable to perform reduction because all keys have been selected. Returning combined object only"
            )
            return (reference.copy(), None) if return_weights else reference.copy()

        query = self[:, rest]

        if mode == Reduction.SCALE:
            reference = _row_normalize(reference)
        elif mode == Reduction.DIST:
            # compute a set of weights of shape (n_query x n_reference)
            if dist_measure == DistanceMeasure.COSINE_SIM:
                weights = _cosine_sim(reference.X, query.X)
            elif dist_measure == DistanceMeasure.WASSERSTEIN_DIST:
                weights = _wasserstein_dist(reference.X, query.X)
            elif dist_measure == DistanceMeasure.KL_DIV:
                weights = _kl_div(reference.X, query.X)
            elif dist_measure == DistanceMeasure.JS_DIV:
                weights = _js_div(reference.X, query.X)
            elif dist_measure == DistanceMeasure.MUTUAL_INFO:
                weights = _mutual_info(reference.X, query.X)
            elif dist_measure == DistanceMeasure.EQUAL:
                weights = _row_normalize(np.ones((query.shape[1], reference.shape[1])))
            else:
                raise NotImplementedError(f"Distance measure `{dist_measure}` is not yet implemented.")

            # make some checks on the weights
            if weights.shape != (query.shape[1], reference.shape[1]):
                raise ValueError(
                    f"Expected weight matrix to be of shape `({query.shape[1]}, {reference.shape[1]})`, "
                    f"found `{weights.shape}`."
                )
            if not np.isfinite(weights).all():
                raise ValueError("Weights matrix contains elements that are not finite.")
            if (weights < 0).any():
                raise ValueError("Weights matrix contains negative elements.")

            if (weights == 0).any():
                logg.warning("Weights matrix contains exact zeros.")

            # normalize the weights to row-sum to one
            if normalize_weights == NormWeights.SCALE:
                weights_n = _row_normalize(weights)
            elif normalize_weights == NormWeights.SOFTMAX:
                weights_n = _softmax(_row_normalize(weights), softmax_scale)
            else:
                raise NotImplementedError(f"Normalization method `{normalize_weights}` is yet implemented.")

            # check that the weights row-sum to one now
            if not np.allclose(weights_n.sum(1), 1.0):
                raise ValueError("Weights do not sum to 1 row-wise.")

            # use the weights to re-distribute probability mass form query to reference
            for i, w in enumerate(weights_n):
                reference += np.dot(query[:, i].X, w[None, :])
        else:
            raise NotImplementedError(f"Reduction mode `{mode}` is not yet implemented.")

        # check that the lineages row-sum to one now
        if not np.allclose(reference.sum(1), 1.0):
            raise ValueError("Reduced lineage rows do not sum to 1.")

        # potentially create a weights-df and return everything
        if return_weights:
            if mode == Reduction.DIST:
                return (
                    reference,
                    pd.DataFrame(data=weights_n, columns=reference.names, index=query.names),
                )
            return reference, None

        return reference

    @classmethod
    @d.dedent
    @inject_docs(lk=LinKind)
    def from_adata(
        cls,
        adata: AnnData,
        backward: bool = False,
        estimator_backward: Optional[bool] = None,
        kind: Literal["macrostates", "term_states", "fate_probs"] = LinKind.FATE_PROBS,
        copy: bool = False,
    ) -> "Lineage":
        """Reconstruct the :class:`~cellrank.Lineage` object from :class:`~anndata.AnnData` object.

        Parameters
        ----------
        %(adata)s
        %(backward)s
        estimator_backward
            Key which helps to determine whether these states are initial or terminal.
        kind
            Which kind of object to reconstruct. Valid options are:

            - ``{lk.MACROSTATES!r}``- macrostates memberships from :class:`cellrank.estimators.GPCCA`.
            - ``{lk.TERM_STATES!r}``- terminal states memberships from :class:`cellrank.estimators.GPCCA`.
            - ``{lk.FATE_PROBS!r}``- fate probabilities.
        copy
            Whether to return a copy of the underlying array.

        Returns
        -------
        The reconstructed lineage object.
        """
        kind = LinKind(kind)
        if kind == LinKind.MACROSTATES:
            nkey = Key.obs.macrostates(backward)
            key = Key.obsm.memberships(nkey)
        elif kind == LinKind.TERM_STATES:
            nkey = Key.obs.term_states(estim_bwd=estimator_backward, bwd=backward)
            key = Key.obsm.memberships(nkey)
        elif kind == LinKind.FATE_PROBS:
            nkey = Key.obs.term_states(estim_bwd=estimator_backward, bwd=backward)
            key = Key.obsm.fate_probs(backward)
        else:
            raise NotImplementedError(f"Lineage kind `{kind}` is not yet implemented.")

        ckey = Key.uns.colors(nkey)

        if key not in adata.obsm:
            raise KeyError(f"Unable to find lineage data in `adata.obsm[{key!r}]`.")
        data: Union[np.ndarray, Lineage] = adata.obsm[key]
        if copy:
            data = copy_.copy(data)
        if isinstance(data, Lineage):
            return data
        if data.ndim != 2:
            raise ValueError(f"Expected 2 dimensional data, found `{data.ndim}`.")

        states = adata.obs.get(nkey, None)
        if states is None:
            logg.warning(f"Unable to find states in `adata.obs[{nkey!r}]`. Using default names")
        elif not isinstance(states.dtype, pd.CategoricalDtype):
            logg.warning(
                f"Expected `adata.obs[{key!r}]` to be `categorical`, "
                f"found `{infer_dtype(adata.obs[nkey])}`. Using default names"
            )
        else:
            states = list(states.cat.categories)
            if len(states) != data.shape[1]:
                logg.warning(
                    f"Expected to find `{data.shape[1]}` names, found `{len(states)}`. " f"Using default names"
                )
        if states is None or len(states) != data.shape[1]:
            states = [str(i) for i in range(data.shape[1])]

        colors = adata.uns.get(ckey, None)
        if colors is None:
            logg.warning(f"Unable to find colors in `adata.uns[{ckey!r}]`. " f"Using default colors")
        elif len(colors) != data.shape[1]:
            logg.warning(f"Expected to find `{data.shape[1]}` colors, found `{len(colors)}`. " f"Using default colors")
            colors = None

        return Lineage(data, names=states, colors=colors)

    def view(self, dtype=None, type=None, *_, **__) -> "LineageView":
        """Return a view of self."""
        return LineageView(self)

    def __repr__(self) -> str:
        return f'{super().__repr__()[:-1]},\n  names([{", ".join(self.names)}]))'

    def __str__(self):
        return f'{super().__str__()}\n names=[{", ".join(self.names)}]'

    @property
    def _fmt(self) -> Callable[[Any], str]:
        return "{:.06f}".format if np.issubdtype(self.dtype, float) else "{}".format

    def _repr_html_(self) -> str:
        def format_row(r):
            rng = (
                range(self.shape[1])
                if not self._is_transposed or (self._is_transposed and self.shape[1] <= _HTML_REPR_THRESH)
                else list(range(_HT_CELLS)) + [...] + list(range(self.shape[1] - _HT_CELLS - 1, self.shape[1] - 1))
            )

            cells = "".join(
                f"<td style='text-align: right;'>" f"{self._fmt(self.X[r, c])}" f"</td>"
                if isinstance(c, int)
                else _DUMMY_CELL
                for c in rng
            )
            return f"<tr>{(names[r] if self._is_transposed else '') + cells}</tr>"

        def dummy_row() -> str:
            values = "".join(_DUMMY_CELL for _ in range(self.shape[1]))
            return f"<tr>{values}</tr>"

        if self.names is None or self.colors is None:
            raise RuntimeError(
                f"Name or colors are `None`. This can happen when running `array.view({type(self).__name__}`."
            )

        if self.ndim != 2:
            return repr(self)

        styles = [
            f"'background-color: {bg}; color: {fg}; text-align: center; word-wrap: break-word; max-width: 100px'"
            for bg, fg in map(_get_bg_fg_colors, self.colors)
        ]
        names = [f"<th style={style}>{n}</th>" for n, style in zip(self.names, styles)]
        header = f"<tr>{''.join(names)}</tr>"

        if self.shape[0] > _HTML_REPR_THRESH:
            body = "".join(format_row(i) for i in range(_HT_CELLS))
            body += dummy_row()
            body += "".join(format_row(i) for i in range(self.shape[0] - _HT_CELLS - 1, self.shape[0] - 1))
        else:
            body = "".join(format_row(i) for i in range(self.shape[0]))

        cells = "cells" if self.shape[0] > 1 else "cell"
        lineages = "lineages" if self.shape[1] > 1 else "lineage"
        if self._is_transposed:
            cells, lineages = lineages, cells
        metadata = f"<p>{self.shape[0]} {cells} x {self.shape[1]} {lineages}</p>"

        if self._is_transposed:
            header = ""
        return (
            f"<div style='scoped' class='rendered_html'>"
            f"<table class='dataframe'>{header}{body}</table>{metadata}"
            f"</div>"
        )

    def __format__(self, format_spec):
        if self.shape == (1, 1):
            return format_spec.format(self.X[0, 0])
        if self.shape == (1,):
            return format_spec.format(self.X[0])
        return NotImplemented

    def __setstate__(self, state, *_, **__):
        *state, names, colors, is_t = state
        names = names[-1]
        colors = colors[-1]

        self._names = np.empty(names[1])
        self._colors = np.empty(colors[1])

        super().__setstate__(tuple(state))
        self._names.__setstate__(tuple(names))
        self._colors.__setstate__(tuple(colors))

        self._is_transposed = is_t
        self._n_lineages = len(self.names)
        self._names_to_ixs = {name: ix for ix, name in enumerate(self.names)}

    def __reduce__(self):
        res = list(super().__reduce__())

        names = self.names.__reduce__()
        colors = self.colors.__reduce__()
        res[-1] += (names, colors, self._is_transposed)

        return tuple(res)

    def copy(self, _="C") -> "Lineage":
        """Return a copy of itself."""
        obj = Lineage(
            self.T if self._is_transposed else self,
            names=np.array(self.names, copy=True, order=_ORDER),
            colors=np.array(self.colors, copy=True, order=_ORDER),
        )
        return obj.T if self._is_transposed else obj

    def __copy__(self):
        return self.copy()

    def _check_axis1_shape(self, array: Iterable[Union[str, ColorLike]], msg: str) -> List[Union[str, ColorLike]]:
        """Check whether the size of the 1D array has the correct length."""
        array = list(array)
        if len(array) != self._n_lineages:
            raise ValueError(msg.format(self._n_lineages, len(array)))

        return array

    def _maybe_convert_names(
        self,
        names: Iterable[Union[int, str, bool]],
        is_singleton: bool = False,
        default: Optional[Union[int, str]] = None,
        make_unique: bool = True,
    ) -> Union[int, List[int], List[bool]]:
        """Convert string indices to their corresponding int indices."""
        from cellrank._utils._utils import _unique_order_preserving

        if all(isinstance(n, (bool, np.bool_)) for n in names):
            return list(names)
        res = []
        for name in names:
            if isinstance(name, str):
                if name in self._names_to_ixs:
                    name = self._names_to_ixs[name]
                elif default is not None:
                    if isinstance(default, str):
                        if default not in self._names_to_ixs:
                            raise KeyError(
                                f"Invalid lineage name: `{name}`. " f"Valid names are: `{list(self.names)}`."
                            )
                        name = self._names_to_ixs[default]
                    else:
                        name = default
                else:
                    raise KeyError(f"Invalid lineage name `{name!r}`. Valid names are: `{list(self.names)}`.")
            res.append(name)

        if make_unique:
            res = _unique_order_preserving(res)

        return res[0] if is_singleton else res

    @staticmethod
    def _prepare_annotation(
        array: List[str],
        checker: Optional[Callable] = None,
        transformer: Optional[Callable] = None,
        checker_msg: Optional[str] = None,
    ) -> np.ndarray:
        if checker is not None:
            assert checker_msg, "Please provide a message when `checker` is not `None`."
            for v in array:
                if not checker(v):
                    raise ValueError(checker_msg.format(v))
        if transformer is not None:
            array = np.array([transformer(v) for v in array])

        return np.array(array)


class LineageView(Lineage):
    """View of :class:`~cellrank.Lineage`."""

    def __new__(cls, lineage: Lineage) -> "LineageView":
        """Create a LineageView."""
        if not isinstance(lineage, Lineage):
            raise TypeError(f"Cannot create a `{cls.__name__}` of `{type(lineage).__name__}`.")

        view = np.array(lineage, copy=False).view(cls)
        view._owner = lineage
        view._names = lineage.names
        view._n_lineages = len(view.names)
        view._names_to_ixs = lineage._names_to_ixs
        view._colors = lineage.colors
        view._is_transposed = lineage._is_transposed

        return view

    @property
    def names(self) -> np.ndarray:
        """Lineage names."""
        return super().names

    @property
    def owner(self) -> Lineage:
        """Return the lineage associated with this view."""
        return self._owner

    @names.setter
    def names(self, _):
        raise RuntimeError(f"Unable to set names of `{type(self).__name__}`.")

    @property
    def colors(self) -> np.ndarray:
        """Lineage colors."""
        return super().colors

    @colors.setter
    def colors(self, _):
        raise RuntimeError(f"Unable to set colors of `{type(self).__name__}`.")

    def view(self, dtype=None, type=None, *_, **__) -> "LineageView":
        """Return self."""
        return self

    def copy(self, _="C") -> Lineage:
        """Return a copy of self."""
        was_trasposed = False
        if self._is_transposed:
            self = self.T
            was_trasposed = True

        obj = Lineage(
            self,
            names=np.array(self.names, copy=True, order=_ORDER),
            colors=np.array(self.colors, copy=True, order=_ORDER),
        )
        return obj.T if was_trasposed else obj


def _remove_zero_rows(a: np.ndarray, b: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    if a.shape[0] != b.shape[0]:
        raise ValueError("Lineage objects have unequal cell numbers")

    bool_a = (a == 0).any(axis=1)
    bool_b = (b == 0).any(axis=1)
    mask = ~np.logical_or(bool_a, bool_b)

    logg.warning(f"Removed {a.shape[0] - np.sum(mask)} rows because they contained zeros")

    return a[mask, :], b[mask, :]


def _softmax(X, beta: float = 1):
    return np.exp(X * beta) / np.expand_dims(np.sum(np.exp(X * beta), axis=1), -1)


def _row_normalize(X: Union[np.ndarray, Lineage]) -> Union[np.ndarray, Lineage]:
    if isinstance(X, Lineage):
        return X / X.sum(1)  # lineage is shape-preserving
    return X / X.sum(1, keepdims=True)


def _col_normalize(X, norm_ord=2):
    from numpy.linalg import norm

    return X / norm(X, ord=norm_ord, axis=0)


def _cosine_sim(reference, query):
    # the cosine similarity is symmetric
    # normalize these to have 2-norm 1
    reference_n, query_n = _col_normalize(reference, 2), _col_normalize(query, 2)
    return (reference_n.T @ query_n).T


def _point_wise_distance(reference, query, distance):
    # utility function for all point-wise distances/divergences
    # take care of rows that contain zeros
    reference_no_zero, query_no_zero = _remove_zero_rows(reference, query)

    # normalize these to be valid probability distributions (column-wise)
    reference_n, query_n = (
        _col_normalize(reference_no_zero, 1),
        _col_normalize(query_no_zero, 1),
    )

    # loop over query and reference columns and compute pairwise wasserstein distances
    weights = np.zeros((query.shape[1], reference.shape[1]))
    for i, q_d in enumerate(query_n.T):
        for j, r_d in enumerate(reference_n.T):
            weights[i, j] = 1.0 / distance(q_d, r_d)

    return weights


def _wasserstein_dist(reference, query):
    # the wasserstein distance is symmetric
    return _point_wise_distance(reference, query, st.wasserstein_distance)


def _kl_div(reference, query):
    return _point_wise_distance(reference, query, st.entropy)


def _js_div(reference, query):
    # the js divergence is symmetric
    from scipy.spatial.distance import jensenshannon

    return _point_wise_distance(reference, query, jensenshannon)


def _mutual_info(reference, query):
    # mutual information is not symmetric. We don't need to normalise the vectors, it's invariant under scaling.
    from sklearn.feature_selection import mutual_info_regression

    weights = np.zeros((query.shape[1], reference.shape[1]))
    for i, target in enumerate(query.T):
        weights[i, :] = mutual_info_regression(reference, target)

    return weights


@_REGISTRY.register_write(H5Group, Lineage, IOSpec("array", "0.2.0"))
@_REGISTRY.register_write(H5Group, LineageView, IOSpec("array", "0.2.0"))
@_REGISTRY.register_write(ZarrGroup, Lineage, IOSpec("array", "0.2.0"))
@_REGISTRY.register_write(ZarrGroup, LineageView, IOSpec("array", "0.2.0"))
def _write_lineage(
    f: Any,
    k: str,
    elem: Union[Lineage, LineageView],
    _writer: Any,
    dataset_kwargs: Mapping[str, Any] = types.MappingProxyType({}),
) -> None:
    write_basic(f, k, elem=elem.X, _writer=_writer, dataset_kwargs=dataset_kwargs)



import contextlib
import pathlib
import pickle
from typing import Optional, Protocol, Tuple, Union

from anndata import AnnData

from cellrank import logging as logg

__all__ = ["IOMixin"]


class IOMixinProtocol(Protocol):
    @property
    def shape(self) -> Tuple[int, ...]:
        ...

    @property
    def adata(self) -> AnnData:
        ...

    @adata.setter
    def adata(self, adata: AnnData) -> None:
        ...


class IOMixin:
    """Mixin that allows for serialization from/to files using :mod:`pickle`."""

    @property
    @contextlib.contextmanager
    def _remove_adata(self) -> None:
        """Temporarily remove :attr:`adata`, if present."""
        adata = getattr(self, "adata", None)

        try:
            if adata is not None:
                self.adata = None
            yield
        finally:
            if adata is not None:
                self.adata = adata

    def write(
        self,
        fname: Union[str, pathlib.Path],
        write_adata: bool = True,
    ) -> None:
        """Serialize self to a file using :mod:`pickle`.

        Parameters
        ----------
        fname
            Path where to save the object.
        write_adata
            Whether to save :attr:`adata` object.

        Returns
        -------
        Nothing, just writes itself to a file.
        """
        logg.info(f"Writing `{self}` to `{fname}`")

        if write_adata:
            with open(fname, "wb") as fout:
                pickle.dump(self, fout)
        else:
            with self._remove_adata, open(fname, "wb") as fout:
                pickle.dump(self, fout)

    @staticmethod
    def read(fname: Union[str, pathlib.Path], adata: Optional[AnnData] = None, copy: bool = False) -> "IOMixin":
        """De-serialize self from a file.

        Parameters
        ----------
        fname
            Path from which to read the object.
        adata
            :class:`~anndata.AnnData` object to assign to the saved object.
            Only used when the saved object has :attr:`adata` and it was saved without it.
        copy
            Whether to copy ``adata`` before assigning it. If ``adata`` is a view, it is always copied.

        Returns
        -------
        The de-serialized object.
        """
        with open(fname, "rb") as fin:
            obj: IOMixinProtocol = pickle.load(fin)

        if hasattr(obj, "adata"):
            if isinstance(obj.adata, AnnData):
                if adata is not None:
                    logg.warning("Ignoring supplied `adata` object because it is already present")
                return obj

            if not isinstance(adata, AnnData):
                raise TypeError(
                    "This object was saved without its `adata` object. " "Please supply one as `adata=...`."
                )

            if obj.shape[0] != len(adata):
                raise ValueError(f"Expected `adata` to be of length `{len(adata)}`, found `{obj.shape[0]}`.")
            if copy or adata.is_view:
                adata = adata.copy()

            obj.adata = adata
            return obj

        return obj



import datetime
import functools
import logging
from typing import Iterable, Optional

__all__ = [
    "print_versions",
    "print_version_and_date",
    "hint",
    "info",
    "warning",
    "error",
    "debug",
]

HINT = (logging.INFO + logging.DEBUG) // 2
logging.addLevelName(HINT, "HINT")


class _RootLogger(logging.RootLogger):
    def __init__(self, level):
        super().__init__(level)
        self.propagate = False
        _RootLogger.manager = logging.Manager(self)

    def log(
        self,
        level: int,
        msg: str,
        *,
        extra: Optional[dict] = None,
        time: datetime.datetime = None,
        deep: Optional[str] = None,
    ) -> datetime.datetime:
        from cellrank import settings

        # this will correctly initialize the handles if doing
        # just from cellrank import logging
        settings.verbosity = settings.verbosity

        now = datetime.datetime.now(datetime.timezone.utc)
        time_passed: datetime.timedelta = None if time is None else now - time
        extra = {
            **(extra or {}),
            "deep": deep if settings.verbosity.level < level else None,
            "time_passed": time_passed,
        }
        _ = super().log(level, msg, extra=extra)
        return now

    def critical(self, msg: str, *, time=None, deep=None, extra=None) -> datetime.datetime:
        return self.log(logging.CRITICAL, msg, time=time, deep=deep, extra=extra)

    def error(self, msg: str, *, time=None, deep=None, extra=None) -> datetime.datetime:
        return self.log(logging.ERROR, msg, time=time, deep=deep, extra=extra)

    def warning(self, msg: str, *, time=None, deep=None, extra=None) -> datetime.datetime:
        return self.log(logging.WARNING, msg, time=time, deep=deep, extra=extra)

    def info(self, msg: str, *, time=None, deep=None, extra=None) -> datetime.datetime:
        return self.log(logging.INFO, msg, time=time, deep=deep, extra=extra)

    def hint(self, msg: str, *, time=None, deep=None, extra=None) -> datetime.datetime:
        return self.log(HINT, msg, time=time, deep=deep, extra=extra)

    def debug(self, msg: str, *, time=None, deep=None, extra=None) -> datetime.datetime:
        return self.log(logging.DEBUG, msg, time=time, deep=deep, extra=extra)


class _LogFormatter(logging.Formatter):
    def __init__(self, fmt="{levelname}: {message}", datefmt="%Y-%m-%d %H:%M", style="{"):
        super().__init__(fmt, datefmt, style)

    def format(self, record: logging.LogRecord):
        format_orig = self._style._fmt
        if record.levelno == logging.INFO:
            self._style._fmt = "{message}"
        elif record.levelno == HINT:
            self._style._fmt = "--> {message}"
        elif record.levelno == logging.DEBUG:
            self._style._fmt = "DEBUG: {message}"
        if record.time_passed:
            # strip microseconds
            if record.time_passed.microseconds:
                record.time_passed = datetime.timedelta(seconds=int(record.time_passed.total_seconds()))
            if "{time_passed}" in record.msg:
                record.msg = record.msg.replace("{time_passed}", str(record.time_passed))
            else:
                self._style._fmt += " ({time_passed})"
        if record.deep:
            record.msg = f"{record.msg}: {record.deep}"
        result = logging.Formatter.format(self, record)
        self._style._fmt = format_orig
        return result


_DEPENDENCIES_NUMERICS = [
    "scanpy",
    "anndata",
    "numpy",
    "numba",
    "scipy",
    "pandas",
    "pygpcca",
    ("sklearn", "scikit-learn"),
    "statsmodels",
    ("igraph", "python-igraph"),
    "scvelo",
    "pygam",
]


_DEPENDENCIES_PLOTTING = ["matplotlib", "seaborn"]


def _versions_dependencies(dependencies: Iterable[str]):
    # this is not the same as the requirements!
    for mod in dependencies:
        mod_name, dist_name = mod if isinstance(mod, tuple) else (mod, mod)
        try:
            imp = __import__(mod_name)
            yield dist_name, imp.__version__
        except (ImportError, AttributeError):
            pass


def print_versions():
    """Print package versions that might influence the numerical and plotting results."""
    from cellrank import settings

    modules = ["cellrank"] + _DEPENDENCIES_NUMERICS + _DEPENDENCIES_PLOTTING
    print(
        " ".join(f"{mod}=={ver}" for mod, ver in _versions_dependencies(modules)),
        file=settings.logfile,
    )


def print_version_and_date():
    """
    Print version and date.

    Useful for starting a notebook so you see when you started working.
    """
    from cellrank import __version__, settings

    print(
        f"Running CellRank {__version__}, on {datetime.datetime.now():%Y-%m-%d %H:%M}.",
        file=settings.logfile,
    )


def _copy_docs_and_signature(fn):
    return functools.partial(functools.update_wrapper, wrapped=fn, assigned=["__doc__", "__annotations__"])


def error(
    msg: str,
    *,
    time: datetime.datetime = None,
    deep: Optional[str] = None,
    extra: Optional[dict] = None,
) -> datetime.datetime:
    """
    Log message with specific level and return current time.

    Parameters
    ----------
    msg
        Message to display.
    time
        A time in the past. If this is passed, the time difference from then
        to now is appended to `msg` as ` (HH:MM:SS)`.
        If `msg` contains `{time_passed}`, the time difference is instead
        inserted at that position.
    deep
        If the current verbosity is higher than the log function’s level,
        this gets displayed as well
    extra
        Additional values you can specify in `msg` like `{time_passed}`.

    Returns
    -------
    The current time.
    """
    from cellrank import settings

    return settings._root_logger.error(msg, time=time, deep=deep, extra=extra)


@_copy_docs_and_signature(error)
def warning(msg: str, *, time=None, deep=None, extra=None) -> datetime.datetime:
    from cellrank import settings

    return settings._root_logger.warning(msg, time=time, deep=deep, extra=extra)


@_copy_docs_and_signature(error)
def info(msg: str, *, time=None, deep=None, extra=None) -> datetime.datetime:
    from cellrank import settings

    return settings._root_logger.info(msg, time=time, deep=deep, extra=extra)


@_copy_docs_and_signature(error)
def hint(msg: str, *, time=None, deep=None, extra=None) -> datetime.datetime:
    from cellrank import settings

    return settings._root_logger.hint(msg, time=time, deep=deep, extra=extra)


@_copy_docs_and_signature(error)
def debug(msg: str, *, time=None, deep=None, extra=None) -> datetime.datetime:
    from cellrank import settings

    return settings._root_logger.debug(msg, time=time, deep=deep, extra=extra)
